"""Test cases for shredder mitigation."""

import os
from datetime import date, datetime, time
from pathlib import Path
from unittest.mock import call, patch

import click
import pytest
import yaml
from click.exceptions import ClickException
from click.testing import CliRunner
from gcloud import bigquery  # type: ignore

from bigquery_etl.backfill.shredder_mitigation import (
    PREVIOUS_DATE,
    SHREDDER_MITIGATION_CHECKS_NAME,
    SHREDDER_MITIGATION_QUERY_NAME,
    Column,
    ColumnStatus,
    ColumnType,
    DataTypeGroup,
    Subset,
    classify_columns,
    generate_query_with_shredder_mitigation,
    get_bigquery_type,
    validate_types,
)


class TestClassifyColumns:
    """Test cases for function classify_columns."""

    def test_new_numeric_dimension(self):
        """Test adding a numeric dimension."""
        new_row = {
            "submission_date": "2024-01-01",
            "app_name": "Browser",
            "channel": "Beta",
            "first_seen_year": 2024,
            "os_build": "Windows 10",
            "metric_numeric": 10,
            "metric_float": 15.123456,
        }
        existing_columns = [
            "submission_date",
            "app_name",
            "channel",
            "first_seen_year",
            "os",
        ]
        new_columns = [
            "submission_date",
            "app_name",
            "channel",
            "first_seen_year",
            "os_build",
        ]
        existing_schema = [
            bigquery.SchemaField("submission_date", "DATE", "NULLABLE", None, None),
            bigquery.SchemaField("app_name", "STRING", "NULLABLE", None, None),
            bigquery.SchemaField("channel", "STRING", "NULLABLE", None, None),
            bigquery.SchemaField("first_seen_year", "INT", "NULLABLE", None, None),
            bigquery.SchemaField("os", "STRING", "NULLABLE", None, None),
            bigquery.SchemaField("metric_numeric", "NUMERIC", "NULLABLE", None, None),
        ]
        new_schema = [
            bigquery.SchemaField("submission_date", "DATE", "NULLABLE", None, None),
            bigquery.SchemaField("app_name", "STRING", "NULLABLE", None, None),
            bigquery.SchemaField("channel", "STRING", "NULLABLE", None, None),
            bigquery.SchemaField("first_seen_year", "INT", "NULLABLE", None, None),
            bigquery.SchemaField("os_build", "STRING", "NULLABLE", None, None),
            bigquery.SchemaField("metric_numeric", "NUMERIC", "NULLABLE", None, None),
            bigquery.SchemaField("metric_float", "FLOAT", "NULLABLE", None, None),
        ]

        expected_common_dimensions = [
            Column(
                name="app_name",
                data_type=DataTypeGroup.STRING,
                column_type=ColumnType.DIMENSION,
                status=ColumnStatus.COMMON,
            ),
            Column(
                name="channel",
                data_type=DataTypeGroup.STRING,
                column_type=ColumnType.DIMENSION,
                status=ColumnStatus.COMMON,
            ),
            Column(
                name="first_seen_year",
                data_type=DataTypeGroup.INTEGER,
                column_type=ColumnType.DIMENSION,
                status=ColumnStatus.COMMON,
            ),
            Column(
                name="submission_date",
                data_type=DataTypeGroup.DATE,
                column_type=ColumnType.DIMENSION,
                status=ColumnStatus.COMMON,
            ),
        ]

        expected_added_dimensions = [
            Column(
                name="os_build",
                data_type=DataTypeGroup.STRING,
                column_type=ColumnType.DIMENSION,
                status=ColumnStatus.ADDED,
            )
        ]

        expected_removed_dimensions = [
            Column(
                name="os",
                data_type=DataTypeGroup.STRING,
                column_type=ColumnType.UNDETERMINED,
                status=ColumnStatus.REMOVED,
            )
        ]

        expected_metrics = [
            Column(
                name="metric_float",
                data_type=DataTypeGroup.FLOAT,
                column_type=ColumnType.METRIC,
                status=ColumnStatus.COMMON,
            ),
            Column(
                name="metric_numeric",
                data_type=DataTypeGroup.NUMERIC,
                column_type=ColumnType.METRIC,
                status=ColumnStatus.COMMON,
            ),
        ]

        common_dimensions, added_dimensions, removed_dimensions, metrics, undefined = (
            classify_columns(
                new_row, existing_columns, new_columns, existing_schema, new_schema
            )
        )

        assert common_dimensions == expected_common_dimensions
        assert added_dimensions == expected_added_dimensions
        assert removed_dimensions == expected_removed_dimensions
        assert metrics == expected_metrics
        assert undefined == []

    def test_new_boolean_dimension(self):
        """Test adding a boolean dimension."""
        new_row = {
            "submission_date": "2024-01-01",
            "metric_float": 10.000,
            "channel": "Beta",
            "first_seen_year": 2024,
            "is_default_browser": False,
            "metric_numeric": 10,
        }
        existing_columns = ["submission_date", "channel", "first_seen_year"]
        new_columns = [
            "submission_date",
            "channel",
            "first_seen_year",
            "is_default_browser",
        ]
        existing_schema = [
            bigquery.SchemaField("submission_date", "DATE", "NULLABLE", None, None),
            bigquery.SchemaField("channel", "STRING", "NULLABLE", None, None),
            bigquery.SchemaField("first_seen_year", "INT", "NULLABLE", None, None),
            bigquery.SchemaField("metric_numeric", "NUMERIC", "NULLABLE", None, None),
            bigquery.SchemaField("metric_float", "FLOAT", "NULLABLE", None, None),
        ]
        new_schema = [
            bigquery.SchemaField("submission_date", "DATE", "NULLABLE", None, None),
            bigquery.SchemaField("channel", "STRING", "NULLABLE", None, None),
            bigquery.SchemaField("first_seen_year", "INT", "NULLABLE", None, None),
            bigquery.SchemaField(
                "is_default_browser", "BOOLEAN", "NULLABLE", None, None
            ),
            bigquery.SchemaField("metric_numeric", "NUMERIC", "NULLABLE", None, None),
            bigquery.SchemaField("metric_float", "FLOAT", "NULLABLE", None, None),
        ]

        expected_common_dimensions = [
            Column(
                name="channel",
                data_type=DataTypeGroup.STRING,
                column_type=ColumnType.DIMENSION,
                status=ColumnStatus.COMMON,
            ),
            Column(
                name="first_seen_year",
                data_type=DataTypeGroup.INTEGER,
                column_type=ColumnType.DIMENSION,
                status=ColumnStatus.COMMON,
            ),
            Column(
                name="submission_date",
                data_type=DataTypeGroup.DATE,
                column_type=ColumnType.DIMENSION,
                status=ColumnStatus.COMMON,
            ),
        ]

        expected_added_dimensions = [
            Column(
                name="is_default_browser",
                data_type=DataTypeGroup.BOOLEAN,
                column_type=ColumnType.DIMENSION,
                status=ColumnStatus.ADDED,
            )
        ]

        expected_removed_dimensions = []

        expected_metrics = [
            Column(
                name="metric_float",
                data_type=DataTypeGroup.FLOAT,
                column_type=ColumnType.METRIC,
                status=ColumnStatus.COMMON,
            ),
            Column(
                name="metric_numeric",
                data_type=DataTypeGroup.NUMERIC,
                column_type=ColumnType.METRIC,
                status=ColumnStatus.COMMON,
            ),
        ]

        common_dimensions, added_dimensions, removed_dimensions, metrics, undefined = (
            classify_columns(
                new_row, existing_columns, new_columns, existing_schema, new_schema
            )
        )

        assert common_dimensions == expected_common_dimensions
        assert added_dimensions == expected_added_dimensions
        assert removed_dimensions == expected_removed_dimensions
        assert metrics == expected_metrics
        assert undefined == []

    def test_removed_numeric_dimension(self):
        """Test removing numeric dimension and metric_integer."""
        new_row = {
            "submission_date": "2024-01-01",
            "metric_float": 1.00003496056549640598605498605486,
            "channel": "Beta",
            "is_default_browser": False,
            "metric_numeric": 101927498327449035043865,
        }
        existing_columns = [
            "submission_date",
            "channel",
            "first_seen_year",
            "is_default_browser",
        ]
        new_columns = ["submission_date", "channel", "is_default_browser"]
        existing_schema = [
            bigquery.SchemaField("submission_date", "DATE", "NULLABLE", None, None),
            bigquery.SchemaField("channel", "STRING", "NULLABLE", None, None),
            bigquery.SchemaField("first_seen_year", "INT", "NULLABLE", None, None),
            bigquery.SchemaField(
                "is_default_browser", "BOOLEAN", "NULLABLE", None, None
            ),
            bigquery.SchemaField("metric_integer", "INTEGER", "NULLABLE", None, None),
            bigquery.SchemaField("metric_float", "FLOAT", "NULLABLE", None, None),
        ]
        new_schema = [
            bigquery.SchemaField("submission_date", "DATE", "NULLABLE", None, None),
            bigquery.SchemaField("channel", "STRING", "NULLABLE", None, None),
            bigquery.SchemaField(
                "is_default_browser", "BOOLEAN", "NULLABLE", None, None
            ),
            bigquery.SchemaField("metric_numeric", "NUMERIC", "NULLABLE", None, None),
            bigquery.SchemaField("metric_float", "FLOAT", "NULLABLE", None, None),
        ]

        expected_common_dimensions = [
            Column(
                name="channel",
                data_type=DataTypeGroup.STRING,
                column_type=ColumnType.DIMENSION,
                status=ColumnStatus.COMMON,
            ),
            Column(
                name="is_default_browser",
                data_type=DataTypeGroup.BOOLEAN,
                column_type=ColumnType.DIMENSION,
                status=ColumnStatus.COMMON,
            ),
            Column(
                name="submission_date",
                data_type=DataTypeGroup.DATE,
                column_type=ColumnType.DIMENSION,
                status=ColumnStatus.COMMON,
            ),
        ]

        expected_added_dimensions = []

        expected_removed_dimensions = [
            Column(
                name="first_seen_year",
                data_type=DataTypeGroup.INTEGER,
                column_type=ColumnType.UNDETERMINED,
                status=ColumnStatus.REMOVED,
            )
        ]

        expected_metrics = [
            Column(
                name="metric_float",
                data_type=DataTypeGroup.FLOAT,
                column_type=ColumnType.METRIC,
                status=ColumnStatus.COMMON,
            ),
            Column(
                name="metric_numeric",
                data_type=DataTypeGroup.NUMERIC,
                column_type=ColumnType.METRIC,
                status=ColumnStatus.COMMON,
            ),
        ]

        common_dimensions, added_dimensions, removed_dimensions, metrics, undefined = (
            classify_columns(
                new_row, existing_columns, new_columns, existing_schema, new_schema
            )
        )

        assert common_dimensions == expected_common_dimensions
        assert added_dimensions == expected_added_dimensions
        assert removed_dimensions == expected_removed_dimensions
        assert metrics == expected_metrics
        assert undefined == []

    def test_new_multiple_dimensions_including_from_values_none(self):
        """Test adding multiple dimensions."""
        new_row = {
            "submission_date": "2024-01-01",
            "os_version": "10.1",
            "metric_float": 1.00003496056549640598605498605486,
            "channel": "Beta",
            "is_default_browser": False,
            "os_version_build": None,
            "segment": None,
        }
        existing_columns = [
            "submission_date",
            "channel",
            "first_seen_year",
            "is_default_browser",
        ]
        new_columns = [
            "submission_date",
            "channel",
            "is_default_browser",
            "os_version",
            "os_version_build",
            "segment",
        ]
        existing_schema = [
            bigquery.SchemaField("submission_date", "DATE", "NULLABLE", None, None),
            bigquery.SchemaField("channel", "STRING", "NULLABLE", None, None),
            bigquery.SchemaField("first_seen_year", "INT", "NULLABLE", None, None),
            bigquery.SchemaField(
                "is_default_browser", "BOOLEAN", "NULLABLE", None, None
            ),
            bigquery.SchemaField("metric_integer", "INTEGER", "NULLABLE", None, None),
            bigquery.SchemaField("metric_float", "FLOAT", "NULLABLE", None, None),
        ]
        new_schema = [
            bigquery.SchemaField("submission_date", "DATE", "NULLABLE", None, None),
            bigquery.SchemaField("channel", "STRING", "NULLABLE", None, None),
            bigquery.SchemaField(
                "is_default_browser", "BOOLEAN", "NULLABLE", None, None
            ),
            bigquery.SchemaField("os_version", "STRING", "NULLABLE", None, None),
            bigquery.SchemaField("os_version_build", "STRING", "NULLABLE", None, None),
            bigquery.SchemaField("segment", "STRING", "NULLABLE", None, None),
            bigquery.SchemaField("metric_numeric", "NUMERIC", "NULLABLE", None, None),
            bigquery.SchemaField("metric_float", "FLOAT", "NULLABLE", None, None),
        ]

        expected_common_dimensions = [
            Column(
                name="channel",
                data_type=DataTypeGroup.STRING,
                column_type=ColumnType.DIMENSION,
                status=ColumnStatus.COMMON,
            ),
            Column(
                name="is_default_browser",
                data_type=DataTypeGroup.BOOLEAN,
                column_type=ColumnType.DIMENSION,
                status=ColumnStatus.COMMON,
            ),
            Column(
                name="submission_date",
                data_type=DataTypeGroup.DATE,
                column_type=ColumnType.DIMENSION,
                status=ColumnStatus.COMMON,
            ),
        ]

        expected_added_dimensions = [
            Column(
                name="os_version",
                data_type=DataTypeGroup.STRING,
                column_type=ColumnType.DIMENSION,
                status=ColumnStatus.ADDED,
            ),
            Column(
                name="os_version_build",
                data_type=DataTypeGroup.STRING,
                column_type=ColumnType.DIMENSION,
                status=ColumnStatus.ADDED,
            ),
            Column(
                name="segment",
                data_type=DataTypeGroup.STRING,
                column_type=ColumnType.DIMENSION,
                status=ColumnStatus.ADDED,
            ),
        ]

        expected_removed_dimensions = [
            Column(
                name="first_seen_year",
                data_type=DataTypeGroup.INTEGER,
                column_type=ColumnType.UNDETERMINED,
                status=ColumnStatus.REMOVED,
            )
        ]

        expected_metrics = [
            Column(
                name="metric_float",
                data_type=DataTypeGroup.FLOAT,
                column_type=ColumnType.METRIC,
                status=ColumnStatus.COMMON,
            )
        ]

        common_dimensions, added_dimensions, removed_dimensions, metrics, undefined = (
            classify_columns(
                new_row, existing_columns, new_columns, existing_schema, new_schema
            )
        )

        assert common_dimensions == expected_common_dimensions
        assert added_dimensions == expected_added_dimensions
        assert removed_dimensions == expected_removed_dimensions
        assert metrics == expected_metrics
        assert undefined == []

    def test_new_metrics(self):
        """Test that metrics added in new version are identified correctly."""
        new_row = {
            "submission_date": "2024-01-01",
            "metric_float": 10000349605654964059860549860.520397593485486,
            "channel": None,
            "metric_bigint": 10000349605654964059860549860520397593485486,
            "metric_numeric": 10000349605654964059860549860.520397593485486,
        }
        existing_columns = ["submission_date", "channel"]
        new_columns = ["submission_date", "channel"]
        existing_schema = [
            bigquery.SchemaField("submission_date", "DATE", "NULLABLE", None, None),
            bigquery.SchemaField("channel", "STRING", "NULLABLE", None, None),
            bigquery.SchemaField("metric_float", "FLOAT", "NULLABLE", None, None),
            bigquery.SchemaField("metric_numeric", "NUMERIC", "NULLABLE", None, None),
            bigquery.SchemaField("metric_bigint", "INTEGER", "NULLABLE", None, None),
        ]
        new_schema = [
            bigquery.SchemaField("submission_date", "DATE", "NULLABLE", None, None),
            bigquery.SchemaField("channel", "STRING", "NULLABLE", None, None),
            bigquery.SchemaField("metric_float", "FLOAT", "NULLABLE", None, None),
            bigquery.SchemaField("metric_numeric", "NUMERIC", "NULLABLE", None, None),
            bigquery.SchemaField("metric_bigint", "INTEGER", "NULLABLE", None, None),
        ]

        expected_common_dimensions = [
            Column(
                name="channel",
                data_type=DataTypeGroup.STRING,
                column_type=ColumnType.DIMENSION,
                status=ColumnStatus.COMMON,
            ),
            Column(
                name="submission_date",
                data_type=DataTypeGroup.DATE,
                column_type=ColumnType.DIMENSION,
                status=ColumnStatus.COMMON,
            ),
        ]

        expected_added_dimensions = []

        expected_removed_dimensions = []

        expected_metrics = [
            Column(
                name="metric_bigint",
                data_type=DataTypeGroup.INTEGER,
                column_type=ColumnType.METRIC,
                status=ColumnStatus.COMMON,
            ),
            Column(
                name="metric_float",
                data_type=DataTypeGroup.FLOAT,
                column_type=ColumnType.METRIC,
                status=ColumnStatus.COMMON,
            ),
            Column(
                name="metric_numeric",
                data_type=DataTypeGroup.NUMERIC,
                column_type=ColumnType.METRIC,
                status=ColumnStatus.COMMON,
            ),
        ]

        common_dimensions, added_dimensions, removed_dimensions, metrics, undefined = (
            classify_columns(
                new_row, existing_columns, new_columns, existing_schema, new_schema
            )
        )
        assert common_dimensions == expected_common_dimensions
        assert added_dimensions == expected_added_dimensions
        assert removed_dimensions == expected_removed_dimensions
        assert metrics == expected_metrics
        assert undefined == []

    def test_not_matching_new_row_and_new_columns(self):
        """Test output when data in bigquery doesn't match columns in group by."""
        new_row = {
            "submission_date": "2024-01-01",
            "channel": None,
            "os": "Mac",
            "metric_int": 2024,
        }
        existing_columns = ["submission_date", "channel"]
        new_columns = ["submission_date", "channel", "os", "is_default_browser"]
        existing_schema = [
            bigquery.SchemaField("submission_date", "DATE", "NULLABLE", None, None),
            bigquery.SchemaField("channel", "STRING", "NULLABLE", None, None),
            bigquery.SchemaField("metric_float", "FLOAT", "NULLABLE", None, None),
        ]
        new_schema = [
            bigquery.SchemaField("submission_date", "DATE", "NULLABLE", None, None),
            bigquery.SchemaField("channel", "STRING", "NULLABLE", None, None),
            bigquery.SchemaField("os", "STRING", "NULLABLE", None, None),
            bigquery.SchemaField(
                "is_default_browser", "BOOLEAN", "NULLABLE", None, None
            ),
            bigquery.SchemaField("metric_float", "FLOAT", "NULLABLE", None, None),
        ]
        expected_exception_text = (
            "Existing dimensions don't match columns retrieved by query."
            " Missing ['is_default_browser']."
        )
        with pytest.raises(ClickException) as e:
            classify_columns(
                new_row, existing_columns, new_columns, existing_schema, new_schema
            )
        assert (str(e.value.message)) == expected_exception_text

    def test_missing_parameters(self):
        """Test output when the parameters to classify columns are incomplete."""
        new_row = {}
        expected_exception_text = (
            f"\n\nMissing one or more required parameters. Received:\nnew_row= {new_row}\n"
            f"existing_dimension_columns= [],\nnew_dimension_columns= []."
        )
        with pytest.raises(ClickException) as e:
            classify_columns(new_row, [], [], None, None)
        assert (str(e.value.message)) == expected_exception_text

        new_row = {"column_1": "2024-01-01", "column_2": "Windows"}
        new_columns = {"column_2"}
        expected_exception_text = (
            f"\n\nMissing one or more required parameters. Received:\nnew_row= {new_row}\n"
            f"existing_dimension_columns= [],\nnew_dimension_columns= {new_columns}."
        )
        with pytest.raises(ClickException) as e:
            classify_columns(new_row, [], new_columns, None, None)
        assert (str(e.value.message)) == expected_exception_text

        new_row = {"column_1": "2024-01-01", "column_2": "Windows"}
        existing_columns = ["column_1"]
        existing_schema = [
            bigquery.SchemaField("column1", "DATE", "NULLABLE", None, None)
        ]
        new_schema = [
            bigquery.SchemaField("column1", "DATE", "NULLABLE", None, None),
            bigquery.SchemaField("column2", "STRING", "NULLABLE", None, None),
        ]
        expected_exception_text = (
            f"\n\nMissing one or more required parameters. Received:\nnew_row= {new_row}\n"
            f"existing_dimension_columns= {existing_columns},\nnew_dimension_columns= []."
        )
        with pytest.raises(ClickException) as e:
            classify_columns(new_row, existing_columns, [], existing_schema, new_schema)
        assert (str(e.value.message)) == expected_exception_text


class TestGetBigqueryType:
    """Test cases for function get_bigquery_type"""

    def test_numeric_group(self):
        """Test function get_bigquery_type with numeric types."""
        assert get_bigquery_type(3) == DataTypeGroup.INTEGER
        assert get_bigquery_type(2024) == DataTypeGroup.INTEGER
        assert get_bigquery_type(9223372036854775807) == DataTypeGroup.INTEGER
        assert get_bigquery_type(123456) == DataTypeGroup.INTEGER
        assert get_bigquery_type(-123456) == DataTypeGroup.INTEGER
        assert get_bigquery_type(789.01) == DataTypeGroup.FLOAT
        assert get_bigquery_type(1.00000000000000000000000456) == DataTypeGroup.FLOAT
        assert get_bigquery_type(999999999999999999999.999999999) == DataTypeGroup.FLOAT
        assert get_bigquery_type(-1.23456) == DataTypeGroup.FLOAT
        assert get_bigquery_type(100000000000000000000.123456789) == DataTypeGroup.FLOAT

    def test_boolean_group(self):
        """Test function get_bigquery_type with boolean types."""
        assert get_bigquery_type(False) == DataTypeGroup.BOOLEAN
        assert get_bigquery_type(True) == DataTypeGroup.BOOLEAN

    def test_date_type(self):
        """Test function get_bigquery_type with datetime types."""
        assert get_bigquery_type("2024-01-01") == DataTypeGroup.DATE
        assert get_bigquery_type("2024-01-01T10:00:00") == DataTypeGroup.DATETIME
        assert get_bigquery_type("2024-01-01T10:00:00Z") == DataTypeGroup.TIMESTAMP
        assert get_bigquery_type("2024-01-01T10:00:00") == DataTypeGroup.DATETIME
        assert get_bigquery_type("2024-08-01 12:34:56 UTC") == DataTypeGroup.TIMESTAMP
        assert get_bigquery_type("2024-09-02 14:30:45") == DataTypeGroup.DATETIME
        assert get_bigquery_type("12:34:56") == DataTypeGroup.TIME
        assert get_bigquery_type(time(12, 34, 56)) == DataTypeGroup.TIME
        assert get_bigquery_type(datetime(2024, 12, 26)) == DataTypeGroup.DATETIME
        assert get_bigquery_type(date(2024, 12, 26)) == DataTypeGroup.DATE

    def test_other_types(self):
        """Test function get_bigquery_type with other types."""
        assert get_bigquery_type("2024") == DataTypeGroup.STRING
        assert get_bigquery_type(None) == DataTypeGroup.UNDETERMINED


class TestValidateTypes:
    """Test cases for function ValidateTypes"""

    columns = ["column1", "column2", "column3"]
    schema = [
        bigquery.SchemaField("column1", "STRING", "NULLABLE", None, None),
        bigquery.SchemaField("column2", "STRING", "NULLABLE", None, None),
        bigquery.SchemaField("column3", "NUMERIC", "NULLABLE", None, None),
    ]

    def test_validate_types_match(self):
        """ "Test that types match the expected result."""
        sample_row = {"column1": "abcd", "column2": "wxyz", "column3": 1234.105}
        result = validate_types(self.columns, self.schema, sample_row)

        assert result == {
            "column1": DataTypeGroup.STRING,
            "column2": DataTypeGroup.STRING,
            "column3": DataTypeGroup.NUMERIC,
        }

    def test_validate_types_mismatch(self):
        """ "Test that type in the schema is returned even if sample row returns a different type."""
        sample_row = {"column1": "abcd", "column2": "wxyz", "column3": "I am a STRING"}
        result = validate_types(self.columns, self.schema, sample_row)

        assert result == {
            "column1": DataTypeGroup.STRING,
            "column2": DataTypeGroup.STRING,
            "column3": DataTypeGroup.NUMERIC,
        }

    def test_validate_types_missing_column_in_schema(self):
        """Test function get_bigquery_type when columns are not found in schema file."""
        names = ["column1", "column2", "column_not_in_schema"]
        schema = [
            bigquery.SchemaField("column1", "STRING", "NULLABLE", None),
            bigquery.SchemaField("column2", "STRING", "NULLABLE", None),
        ]
        sample_row = {
            "column1": "abcd",
            "column2": "wxyz",
            "column_not_in_schema": 1234,
        }
        result = validate_types(names, schema, sample_row)

        assert result == {
            "column1": DataTypeGroup.STRING,
            "column2": DataTypeGroup.STRING,
            "column_not_in_schema": DataTypeGroup.INTEGER,
        }

    def test_validate_types_missing_data(self):
        """Test function get_bigquery_type when columns are not found in bigquery data."""
        sample_row = {"column1": "abcd", "column2": None}
        result = validate_types(self.columns, self.schema, sample_row)

        assert result == {
            "column1": DataTypeGroup.STRING,
            "column2": DataTypeGroup.STRING,
            "column3": DataTypeGroup.NUMERIC,
        }


class TestSubset:
    """Test cases for the methods in class Subset."""

    project_id = "moz-fx-data-shared-prod"
    dataset = "test"
    destination_table = "test_query_v2"
    destination_table_previous = "test_query_v1"
    path = Path("sql") / project_id / dataset / destination_table
    path_previous = Path("sql") / project_id / dataset / destination_table_previous

    @pytest.fixture
    def runner(self):
        return CliRunner()

    @patch("google.cloud.bigquery.Client")
    def test_version(self, mock_client):
        """Test valid version in destinatino table."""
        test_tables_correct = [
            ("test_v1", 1),
            ("test_v10", 10),
            ("test_v0", 0),
        ]
        for table, expected in test_tables_correct:
            test_subset = Subset(
                mock_client, table, None, self.dataset, self.project_id, None
            )
            assert test_subset.version == expected

        test_tables_incorrect = [
            ("test_v-19", 1),
            ("test_v", 1),
            ("test_3", None),
            ("test_10", None),
            ("test_3", 1),
        ]
        for table, expected in test_tables_incorrect:
            test_subset = Subset(
                mock_client, table, None, self.dataset, self.project_id, None
            )
            with pytest.raises(click.ClickException) as e:
                _ = test_subset.version
            assert e.type == click.ClickException
            assert (
                e.value.message
                == f"Invalid or missing table version in {test_subset.destination_table}."
            )

    @patch("google.cloud.bigquery.Client")
    def test_destination_table_not_given(self, mock_client):
        """Test valid version in destination table."""
        test_destination_table = ""

        with pytest.raises(click.ClickException) as e:
            Subset(
                mock_client,
                test_destination_table,
                None,
                self.dataset,
                self.project_id,
                None,
            )
        assert str(e.value.message) == (
            "destination_table not given and it's required to continue."
        )

    @patch("google.cloud.bigquery.Client")
    def test_partitioning(self, mock_client, runner):
        """Test that partitioning type and value associated to a subset are returned as expected."""
        test_subset = Subset(
            mock_client,
            self.destination_table,
            None,
            self.dataset,
            self.project_id,
            None,
        )

        with runner.isolated_filesystem():
            os.makedirs(Path(self.path), exist_ok=True)
            with open(Path(self.path) / "metadata.yaml", "w") as f:
                f.write(
                    "bigquery:\n  time_partitioning:\n    type: day\n    field: submission_date\n"
                    "    require_partition_filter: true"
                )
            assert test_subset.partitioning == {
                "field": "submission_date",
                "type": "DAY",
            }

            with open(Path(self.path) / "metadata.yaml", "w") as f:
                f.write(
                    "bigquery:\n  time_partitioning:\n    type: day\n    "
                    "field: first_seen_date\n    require_partition_filter: true"
                )
            assert test_subset.partitioning == {
                "field": "first_seen_date",
                "type": "DAY",
            }

            with open(Path(self.path) / "metadata.yaml", "w") as f:
                f.write(
                    "friendly_name: ABCD\ndescription: ABCD\nlabels:\n  incremental: true"
                )
            assert test_subset.partitioning == {"field": None, "type": None}

    @patch("google.cloud.bigquery.Client")
    def test_labels(self, mock_client, runner):
        """Test that partitioning type and value associated to a subset are returned as expected."""
        test_subset = Subset(
            mock_client,
            self.destination_table,
            None,
            self.dataset,
            self.project_id,
            None,
        )

        with runner.isolated_filesystem():
            os.makedirs(Path(self.path), exist_ok=True)
            with open(Path(self.path) / "metadata.yaml", "w") as f:
                f.write(
                    "bigquery:\n  time_partitioning:\n    type: day\n    field: submission_date\n"
                    "friendly_name: Test\ndescription: Test\nlabels:\n  change_controlled: true"
                )
            assert "change_controlled" in test_subset.labels

            with open(Path(self.path) / "metadata.yaml", "w") as f:
                f.write(
                    "friendly_name: Test\ndescription: Test\nlabels:\n  "
                    "incremental: true\n  change_controlled: true\n  shredder_mitigation: true"
                )
            for label in ["change_controlled", "shredder_mitigation", "incremental"]:
                assert label in test_subset.labels

            with open(Path(self.path) / "metadata.yaml", "w") as f:
                f.write("friendly_name: Test\ndescription: Test")
            assert test_subset.labels == {}
            assert "shredder_mitigation" not in test_subset.labels

    @patch("google.cloud.bigquery.Client")
    def test_generate_query(self, mock_client):
        """Test method generate_query with expected subset queries and exceptions."""
        test_subset = Subset(
            mock_client,
            self.destination_table,
            None,
            self.dataset,
            self.project_id,
            None,
        )

        test_subset_query = test_subset.generate_query(
            select_list=["column_1"],
            from_clause=f"{self.destination_table_previous}",
            group_by_clause="ALL",
        )
        assert test_subset_query == (
            f"SELECT column_1 FROM {self.destination_table_previous}" f" GROUP BY ALL"
        )

        test_subset_query = test_subset.generate_query(
            select_list=[1, 2, 3],
            from_clause=f"{self.destination_table_previous}",
            order_by_clause="1, 2, 3",
        )
        assert test_subset_query == (
            f"SELECT 1, 2, 3 FROM {self.destination_table_previous}"
            f" ORDER BY 1, 2, 3"
        )

        test_subset_query = test_subset.generate_query(
            select_list=["column_1", 2, 3],
            from_clause=f"{self.destination_table_previous}",
            group_by_clause="1, 2, 3",
        )
        assert test_subset_query == (
            f"SELECT column_1, 2, 3 FROM {self.destination_table_previous}"
            f" GROUP BY 1, 2, 3"
        )

        test_subset_query = test_subset.generate_query(
            select_list=["column_1"], from_clause=f"{self.destination_table_previous}"
        )
        assert (
            test_subset_query
            == f"SELECT column_1 FROM {self.destination_table_previous}"
        )

        test_subset_query = test_subset.generate_query(
            select_list=["column_1"],
            from_clause=f"{self.destination_table_previous}",
            where_clause="column_1 IS NOT NULL",
            group_by_clause="1",
            order_by_clause="1",
        )
        assert test_subset_query == (
            f"SELECT column_1 FROM {self.destination_table_previous}"
            f" WHERE column_1 IS NOT NULL GROUP BY 1 ORDER BY 1"
        )

        test_subset_query = test_subset.generate_query(
            select_list=["column_1"],
            from_clause=f"{self.destination_table_previous}",
            group_by_clause="1",
            order_by_clause="1",
        )
        assert test_subset_query == (
            f"SELECT column_1 FROM {self.destination_table_previous}"
            f" GROUP BY 1 ORDER BY 1"
        )

        test_subset_query = test_subset.generate_query(
            select_list=["column_1"],
            from_clause=f"{self.destination_table_previous}",
            having_clause="column_1 > 1",
        )
        assert (
            test_subset_query
            == f"SELECT column_1 FROM {self.destination_table_previous}"
        )

        test_subset_query = test_subset.generate_query(
            select_list=["column_1"],
            from_clause=f"{self.destination_table_previous}",
            group_by_clause="1",
            having_clause="column_1 > 1",
        )
        assert test_subset_query == (
            f"SELECT column_1 FROM {self.destination_table_previous}"
            f" GROUP BY 1 HAVING column_1 > 1"
        )

        with pytest.raises(ClickException) as e:
            test_subset.generate_query(
                select_list=[],
                from_clause=f"{self.destination_table_previous}",
                group_by_clause="1",
                having_clause="column_1 > 1",
            )
        assert str(e.value.message) == (
            f"Missing required clause to generate query.\n"
            f"Actual: SELECT: [], FROM: {test_subset.full_table_id}"
        )

    @patch("google.cloud.bigquery.Client")
    def test_get_query_path_results(self, mock_client, runner):
        """Test expected results for a mocked BigQuery call."""
        test_subset = Subset(
            mock_client,
            self.destination_table,
            None,
            self.dataset,
            self.project_id,
            None,
        )
        expected = [{"column_1": "1234"}]

        with runner.isolated_filesystem():
            os.makedirs(self.path, exist_ok=True)
            with open(Path(self.path) / "query.sql", "w") as f:
                f.write("SELECT column_1 WHERE submission_date = @submission_date")

            with pytest.raises(FileNotFoundError) as e:
                test_subset.get_query_path_results(None)
                assert "metadata.yaml" in str(e)

            with open(Path(self.path) / "metadata.yaml", "w") as f:
                f.write(
                    "bigquery:\n  time_partitioning:\n    type: day\n    field: submission_date"
                )
            mock_query = mock_client.query
            mock_query.return_value.result.return_value = iter(expected)
            result = test_subset.get_query_path_results(None)
            assert result == expected


class TestGenerateQueryWithShredderMitigation:
    """Test function generate_query_with_shredder_mitigation and returned query for backfill."""

    project_id = "moz-fx-data-shared-prod"
    dataset = "test"
    destination_table = "test_query_v2"
    staging_table_name = f"{project_id}.{dataset}.test_query_v2__2021_01_01"
    destination_table_previous = "test_query_v1"
    path = Path("sql") / project_id / dataset / destination_table
    path_previous = Path("sql") / project_id / dataset / destination_table_previous

    @pytest.fixture
    def runner(self):
        """Runner"""
        return CliRunner()

    @patch("google.cloud.bigquery.Client")
    @patch("bigquery_etl.backfill.shredder_mitigation.classify_columns")
    def test_generate_query_as_expected(
        self, mock_classify_columns, mock_client, runner
    ):
        """Test that query is generated as expected given a set of mock dimensions and metrics."""

        expected = (
            Path("sql") / self.project_id / self.dataset / self.destination_table,
            """-- Query generated using a template for shredder mitigation.
                        WITH new_version AS (
                          SELECT
                            column_1,
                            column_2,
                            metric_1
                          FROM
                            upstream_1
                          GROUP BY
                            column_1,
                            column_2
                        ),
                        new_agg AS (
                          SELECT
                            submission_date,
                            IF(column_1 IS NULL OR column_1 = '??', '???????', column_1) AS column_1,
                            SUM(metric_1) AS metric_1
                          FROM
                            new_version
                          GROUP BY
                            ALL
                        ),
                        previous_agg AS (
                          SELECT
                            submission_date,
                            IF(column_1 IS NULL OR column_1 = '??', '???????', column_1) AS column_1,
                            SUM(metric_1) AS metric_1
                          FROM
                            `moz-fx-data-shared-prod.test.test_query_v1`
                          WHERE
                            submission_date = @submission_date
                          GROUP BY
                            ALL
                        ),
                        shredded AS (
                          SELECT
                            previous_agg.submission_date,
                            previous_agg.column_1,
                            CAST(NULL AS STRING) AS column_2,
                            COALESCE(previous_agg.metric_1, 0) - COALESCE(new_agg.metric_1, 0) AS metric_1
                          FROM
                            previous_agg
                          LEFT JOIN
                            new_agg
                            ON previous_agg.submission_date = new_agg.submission_date
                            AND previous_agg.column_1 = new_agg.column_1
                          WHERE
                            COALESCE(previous_agg.metric_1, 0) > COALESCE(new_agg.metric_1, 0)
                        )
                        SELECT
                          IF(column_1 = '???????', CAST(NULL AS STRING), column_1) AS column_1,
                          IF(column_2 = '???????', CAST(NULL AS STRING), column_2) AS column_2,
                          metric_1
                        FROM
                          new_version
                        UNION ALL
                        SELECT
                          IF(column_1 = '???????', CAST(NULL AS STRING), column_1) AS column_1,
                          IF(column_2 = '???????', CAST(NULL AS STRING), column_2) AS column_2,
                          metric_1
                        FROM
                          shredded""",
        )

        existing_schema = {
            "fields": [
                {"name": "column_1", "type": "DATE", "mode": "NULLABLE"},
                {"name": "metric_1", "type": "INTEGER", "mode": "NULLABLE"},
            ]
        }
        new_schema = {
            "fields": [
                {"name": "column_1", "type": "DATE", "mode": "NULLABLE"},
                {"name": "column_2", "type": "STRING", "mode": "NULLABLE"},
                {"name": "metric_1", "type": "INTEGER", "mode": "NULLABLE"},
            ]
        }

        with runner.isolated_filesystem():
            os.makedirs(self.path, exist_ok=True)
            os.makedirs(self.path_previous, exist_ok=True)
            with open(self.path / "query.sql", "w") as f:
                f.write(
                    "SELECT column_1, column_2, metric_1 FROM upstream_1"
                    " GROUP BY column_1, column_2"
                )
            with open(self.path_previous / "query.sql", "w") as f:
                f.write("SELECT column_1, metric_1 FROM upstream_1 GROUP BY column_1")

            with open(self.path / "schema.yaml", "w") as f:
                f.write(yaml.safe_dump(new_schema))

            with open(self.path_previous / "schema.yaml", "w") as f:
                f.write(yaml.safe_dump(existing_schema))

            with open(Path(self.path) / "metadata.yaml", "w") as f:
                f.write(
                    "bigquery:\n  time_partitioning:\n    type: day\n    "
                    "field: submission_date\n    require_partition_filter: true\n"
                    "labels:\n    shredder_mitigation: true"
                )
            with open(Path(self.path_previous) / "metadata.yaml", "w") as f:
                f.write(
                    "bigquery:\n  time_partitioning:\n    type: day\n    "
                    "field: submission_date\n    require_partition_filter: true\n"
                    "labels:\n    shredder_mitigation: true"
                )

            mock_classify_columns.return_value = (
                [
                    Column(
                        "column_1",
                        DataTypeGroup.STRING,
                        ColumnType.DIMENSION,
                        ColumnStatus.COMMON,
                    )
                ],
                [
                    Column(
                        "column_2",
                        DataTypeGroup.STRING,
                        ColumnType.DIMENSION,
                        ColumnStatus.ADDED,
                    )
                ],
                [],
                [
                    Column(
                        "metric_1",
                        DataTypeGroup.INTEGER,
                        ColumnType.METRIC,
                        ColumnStatus.COMMON,
                    )
                ],
                [],
            )

            with patch.object(
                Subset,
                "get_query_path_results",
                return_value=[{"column_1": "ABC", "column_2": "DEF", "metric_1": 10.0}],
            ):
                assert os.path.isfile(self.path / "query.sql")
                assert os.path.isfile(self.path_previous / "query.sql")
                result = generate_query_with_shredder_mitigation(
                    client=mock_client,
                    project_id=self.project_id,
                    dataset=self.dataset,
                    destination_table=self.destination_table,
                    staging_table_name=self.staging_table_name,
                    backfill_date=PREVIOUS_DATE,
                )
                assert result[0] == expected[0]
                assert result[1] == expected[1].replace("                        ", "")
                assert os.path.isfile(
                    expected[0] / f"{SHREDDER_MITIGATION_QUERY_NAME}.sql"
                )

    @patch("google.cloud.bigquery.Client")
    @patch("bigquery_etl.backfill.shredder_mitigation.classify_columns")
    def test_generate_query_failed_for_missing_partitioning(
        self, mock_classify_columns, mock_client, runner, capfd
    ):
        """Test that function raises exception for required query parameter -partitioning-
        missing in metadata, instead of generating wrong query."""
        existing_schema = {
            "fields": [
                {"name": "column_1", "type": "DATE", "mode": "NULLABLE"},
                {"name": "metric_1", "type": "INTEGER", "mode": "NULLABLE"},
            ]
        }
        new_schema = {
            "fields": [
                {"name": "column_1", "type": "DATE", "mode": "NULLABLE"},
                {"name": "column_2", "type": "STRING", "mode": "NULLABLE"},
                {"name": "metric_1", "type": "INTEGER", "mode": "NULLABLE"},
            ]
        }

        with runner.isolated_filesystem():
            os.makedirs(self.path, exist_ok=True)
            os.makedirs(self.path_previous, exist_ok=True)
            with open(self.path / "query.sql", "w") as f:
                f.write(
                    "SELECT column_1, column_2, metric_1 FROM upstream_1"
                    " WHERE column_1 = @column_1 GROUP BY column_1, column_2"
                )
            with open(self.path_previous / "query.sql", "w") as f:
                f.write(
                    "SELECT column_1, metric_1 FROM upstream_1"
                    " WHERE column_1 = @column_1 GROUP BY column_1"
                )

            with open(self.path / "schema.yaml", "w") as f:
                f.write(yaml.safe_dump(new_schema))

            with open(self.path_previous / "schema.yaml", "w") as f:
                f.write(yaml.safe_dump(existing_schema))

            with open(Path(self.path) / "metadata.yaml", "w") as f:
                f.write("labels:\n    shredder_mitigation: true")
            with open(Path(self.path_previous) / "metadata.yaml", "w") as f:
                f.write(
                    "bigquery:\n  time_partitioning:\n    type: day\n    "
                    "field: submission_date\n    require_partition_filter: true\n"
                    "labels:\n    shredder_mitigation: true"
                )

            mock_classify_columns.return_value = (
                [
                    Column(
                        "column_1",
                        DataTypeGroup.STRING,
                        ColumnType.DIMENSION,
                        ColumnStatus.COMMON,
                    )
                ],
                [
                    Column(
                        "column_2",
                        DataTypeGroup.STRING,
                        ColumnType.DIMENSION,
                        ColumnStatus.ADDED,
                    )
                ],
                [],
                [
                    Column(
                        "metric_1",
                        DataTypeGroup.INTEGER,
                        ColumnType.METRIC,
                        ColumnStatus.COMMON,
                    )
                ],
                [],
            )

            with pytest.raises((TypeError, IndexError)) as e:
                generate_query_with_shredder_mitigation(
                    client=mock_client,
                    project_id=self.project_id,
                    dataset=self.dataset,
                    destination_table=self.destination_table,
                    staging_table_name=self.staging_table_name,
                    backfill_date=PREVIOUS_DATE,
                )

            assert isinstance(e.value, (TypeError, IndexError))
            assert str(e.value) in {
                "'NoneType' object is not iterable",
                "Invalid type",
                "list index out of range",
            }

    @patch("google.cloud.bigquery.Client")
    @patch("bigquery_etl.backfill.shredder_mitigation.classify_columns")
    def test_generate_checks_as_expected(
        self, mock_classify_columns, mock_client, runner
    ):
        """Test that checks are generated as expected when calling the function."""

        expected = (
            Path("sql") / self.project_id / self.dataset / self.destination_table,
            """-- dummy query""",
        )
        expected_checks_content = """-- Checks generated using a template for shredder mitigation.
        -- Rows in previous version not matching in new version. Mismatches can happen when the row is
        -- missing or any column doesn't match, including a single metric difference.

        #fail
        WITH previous AS (
          SELECT
            column_1,
            IF(column_2 IS NULL OR column_2 = '??', '???????', column_2) AS column_2,
            SUM(metric_1) AS metric_1,
            SUM(metric_2) AS metric_2
          FROM
            `moz-fx-data-shared-prod.test.test_query_v1`
          WHERE
            column_1 = @column_1
          GROUP BY
            ALL
        ),
        new_version AS (
          SELECT
            column_1,
            IF(column_2 IS NULL OR column_2 = '??', '???????', column_2) AS column_2,
            SUM(metric_1) AS metric_1,
            SUM(metric_2) AS metric_2
          FROM
            `moz-fx-data-shared-prod.test.test_query_v2__2021_01_01`
          WHERE
            column_1 = @column_1
          GROUP BY
            ALL
        ),
        previous_not_matching AS (
          SELECT
            *
          FROM
            previous
          EXCEPT DISTINCT
          SELECT
            *
          FROM
            new_version
        )
        SELECT
          IF(
            (SELECT COUNT(*) FROM previous_not_matching) > 0,
            ERROR(
                CONCAT(
                  ((SELECT COUNT(*) FROM previous_not_matching)),
                  " rows in the previous data don't match backfilled data! Run auto-generated checks for ",
                  "all mismatches & search for rows missing or with differences in metrics. Sample row in previous version: ",
                  (SELECT TO_JSON_STRING(ARRAY(SELECT AS STRUCT * FROM previous_not_matching LIMIT 1)))
                  )
                ),
            NULL
            );

        -- Rows in new version not matching in previous version. It could be rows added by the process or rows with differences.

        #fail
        WITH previous AS (
          SELECT
            column_1,
            IF(column_2 IS NULL OR column_2 = '??', '???????', column_2) AS column_2,
            SUM(metric_1) AS metric_1,
            SUM(metric_2) AS metric_2
          FROM
            `moz-fx-data-shared-prod.test.test_query_v1`
          WHERE
            column_1 = @column_1
          GROUP BY
            ALL
        ),
        new_version AS (
          SELECT
            column_1,
            IF(column_2 IS NULL OR column_2 = '??', '???????', column_2) AS column_2,
            SUM(metric_1) AS metric_1,
            SUM(metric_2) AS metric_2
          FROM
            `moz-fx-data-shared-prod.test.test_query_v2__2021_01_01`
          WHERE
            column_1 = @column_1
          GROUP BY
            ALL
        ),
        backfilled_not_matching AS (
            SELECT
              *
            FROM
              new_version
            EXCEPT DISTINCT
            SELECT
              *
            FROM
              previous
        )
        SELECT
          IF(
            (SELECT COUNT(*) FROM backfilled_not_matching) > 0,
            ERROR(
              CONCAT(
                ((SELECT COUNT(*) FROM backfilled_not_matching)),
                " rows in backfill don't match previous version of data! Run auto-generated checks for ",
                "all mismatches & search for rows added or with differences in metrics. Sample row in new_version: ",
                (SELECT TO_JSON_STRING(ARRAY(SELECT AS STRUCT * FROM backfilled_not_matching LIMIT 1)))
              )
            ),
            NULL
            );
            """

        existing_schema = {
            "fields": [
                {"name": "column_1", "type": "DATE", "mode": "NULLABLE"},
                {"name": "column_2", "type": "STRING", "mode": "NULLABLE"},
                {"name": "metric_1", "type": "INTEGER", "mode": "NULLABLE"},
                {"name": "metric_2", "type": "NUMERIC", "mode": "NULLABLE"},
            ]
        }
        new_schema = {
            "fields": [
                {"name": "column_1", "type": "DATE", "mode": "NULLABLE"},
                {"name": "column_2", "type": "STRING", "mode": "NULLABLE"},
                {"name": "column_3", "type": "STRING", "mode": "NULLABLE"},
                {"name": "metric_1", "type": "INTEGER", "mode": "NULLABLE"},
                {"name": "metric_2", "type": "NUMERIC", "mode": "NULLABLE"},
            ]
        }

        with runner.isolated_filesystem():
            os.makedirs(self.path, exist_ok=True)
            os.makedirs(self.path_previous, exist_ok=True)
            with open(self.path / "query.sql", "w") as f:
                f.write(
                    "SELECT column_1, column_2, column_3, metric_1, metric_2 FROM upstream_1"
                    " GROUP BY column_1, column_2, column_3"
                )
            with open(self.path_previous / "query.sql", "w") as f:
                f.write(
                    "SELECT column_1, column_2, metric_1, metric_2 FROM upstream_1"
                    " GROUP BY column_1, column_2"
                )

            with open(self.path / "schema.yaml", "w") as f:
                f.write(yaml.safe_dump(new_schema))

            with open(self.path_previous / "schema.yaml", "w") as f:
                f.write(yaml.safe_dump(existing_schema))

            with open(Path(self.path) / "metadata.yaml", "w") as f:
                f.write(
                    "bigquery:\n  time_partitioning:\n    type: day\n    "
                    "field: column_1\n    require_partition_filter: true\n"
                    "labels:\n    shredder_mitigation: true"
                )
            with open(Path(self.path_previous) / "metadata.yaml", "w") as f:
                f.write(
                    "bigquery:\n  time_partitioning:\n    type: day\n    "
                    "field: column_1\n    require_partition_filter: true\n"
                    "labels:\n    shredder_mitigation: true"
                )

            mock_classify_columns.return_value = (
                [
                    Column(
                        "column_1",
                        DataTypeGroup.STRING,
                        ColumnType.DIMENSION,
                        ColumnStatus.COMMON,
                    ),
                    Column(
                        "column_2",
                        DataTypeGroup.STRING,
                        ColumnType.DIMENSION,
                        ColumnStatus.COMMON,
                    ),
                ],
                [
                    Column(
                        "column_3",
                        DataTypeGroup.STRING,
                        ColumnType.DIMENSION,
                        ColumnStatus.ADDED,
                    )
                ],
                [],
                [
                    Column(
                        "metric_1",
                        DataTypeGroup.INTEGER,
                        ColumnType.METRIC,
                        ColumnStatus.COMMON,
                    ),
                    Column(
                        "metric_2",
                        DataTypeGroup.NUMERIC,
                        ColumnType.METRIC,
                        ColumnStatus.COMMON,
                    ),
                ],
                [],
            )

            with patch.object(
                Subset,
                "get_query_path_results",
                return_value=[
                    {
                        "column_1": "2021-01-01",
                        "column_2": "DEF",
                        "column_3": "ABC",
                        "metric_1": 10.0,
                        "metric_2": 1028374.439587349875643,
                    }
                ],
            ):
                assert os.path.isfile(self.path / "query.sql")
                assert os.path.isfile(self.path_previous / "query.sql")
                result = generate_query_with_shredder_mitigation(
                    client=mock_client,
                    project_id=self.project_id,
                    dataset=self.dataset,
                    destination_table=self.destination_table,
                    staging_table_name=self.staging_table_name,
                    backfill_date=PREVIOUS_DATE,
                )
                checks_file = self.path / f"{SHREDDER_MITIGATION_CHECKS_NAME}.sql"
                assert result[0] == expected[0]
                assert os.path.isfile(checks_file)
                with open(checks_file) as file:
                    checks_content = file.read()
                # Normalize multilines to avoid assert failure due to indentation.
                checks_content_normalized = "\n".join(
                    [
                        line.strip()
                        for line in checks_content.splitlines()
                        if line.strip() != ""
                    ]
                ).strip()
                expected_checks_content_normalized = "\n".join(
                    [
                        line.strip()
                        for line in expected_checks_content.splitlines()
                        if line.strip() != ""
                    ]
                ).strip()
                assert checks_content_normalized == expected_checks_content_normalized

    @patch("google.cloud.bigquery.Client")
    def test_missing_previous_version(self, mock_client, runner):
        """Test that the process raises an exception when previous query version is missing."""
        expected_exc = "Extracting GROUP BY from query failed due to sql file or sql text not available."

        with runner.isolated_filesystem():
            os.makedirs(self.path, exist_ok=True)
            with open(Path(self.path) / "query.sql", "w") as f:
                f.write("SELECT column_1, column_2 FROM upstream_1 GROUP BY column_1")
            with open(Path(self.path) / "metadata.yaml", "w") as f:
                f.write(
                    "Friendly name: Test\n" "labels:\n    shredder_mitigation: true"
                )

            with pytest.raises(ClickException) as e:
                generate_query_with_shredder_mitigation(
                    client=mock_client,
                    project_id=self.project_id,
                    dataset=self.dataset,
                    destination_table=self.destination_table,
                    staging_table_name=self.staging_table_name,
                    backfill_date=PREVIOUS_DATE,
                )
            assert (str(e.value.message)) == expected_exc
            assert (e.type) == ClickException

    @patch("google.cloud.bigquery.Client")
    def test_invalid_group_by(self, mock_client, runner):
        """Test that the process raises an exception when the GROUP BY is invalid for any query."""
        expected_exc = (
            "GROUP BY must use an explicit list of columns. "
            "Avoid expressions like `GROUP BY ALL` or `GROUP BY 1, 2, 3`."
        )

        # GROUP BY including a number
        with runner.isolated_filesystem():
            previous_group_by = "column_1, column_2, column_3"
            new_group_by = "3, column_4, column_5"
            os.makedirs(self.path, exist_ok=True)
            os.makedirs(self.path_previous, exist_ok=True)

            with open(Path(self.path) / "query.sql", "w") as f:
                f.write(
                    f"SELECT column_1, column_2 FROM upstream_1 GROUP BY {new_group_by}"
                )
            with open(Path(self.path_previous) / "query.sql", "w") as f:
                f.write(
                    f"SELECT column_1, column_2 FROM upstream_1 GROUP BY {previous_group_by}"
                )
            with open(Path(self.path) / "metadata.yaml", "w") as f:
                f.write(
                    "Friendly name: Test\n" "labels:\n    shredder_mitigation: true"
                )
            with open(Path(self.path_previous) / "metadata.yaml", "w") as f:
                f.write(
                    "Friendly name: Test\n" "labels:\n    shredder_mitigation: true"
                )

            with pytest.raises(ClickException) as e:
                generate_query_with_shredder_mitigation(
                    client=mock_client,
                    project_id=self.project_id,
                    dataset=self.dataset,
                    destination_table=self.destination_table,
                    staging_table_name=self.staging_table_name,
                    backfill_date=PREVIOUS_DATE,
                )
            assert (str(e.value.message)) == expected_exc

            # GROUP BY 1, 2, 3
            previous_group_by = "1, 2, 3"
            new_group_by = "column_1, column_2, column_3"
            with open(Path(self.path) / "query.sql", "w") as f:
                f.write(
                    f"SELECT column_1, column_2 FROM upstream_1 GROUP BY {new_group_by}"
                )
            with open(Path(self.path_previous) / "query.sql", "w") as f:
                f.write(
                    f"SELECT column_1, column_2 FROM upstream_1 GROUP BY {previous_group_by}"
                )
            with pytest.raises(ClickException) as e:
                generate_query_with_shredder_mitigation(
                    client=mock_client,
                    project_id=self.project_id,
                    dataset=self.dataset,
                    destination_table=self.destination_table,
                    staging_table_name=self.staging_table_name,
                    backfill_date=PREVIOUS_DATE,
                )
            assert (str(e.value.message)) == expected_exc

            # GROUP BY ALL
            previous_group_by = "column_1, column_2, column_3"
            new_group_by = "ALL"
            with open(Path(self.path) / "query.sql", "w") as f:
                f.write(
                    f"SELECT column_1, column_2 FROM upstream_1 GROUP BY {new_group_by}"
                )
            with open(Path(self.path_previous) / "query.sql", "w") as f:
                f.write(
                    f"SELECT column_1, column_2 FROM upstream_1 GROUP BY {previous_group_by}"
                )
            with pytest.raises(ClickException) as e:
                generate_query_with_shredder_mitigation(
                    client=mock_client,
                    project_id=self.project_id,
                    dataset=self.dataset,
                    destination_table=self.destination_table,
                    staging_table_name=self.staging_table_name,
                    backfill_date=PREVIOUS_DATE,
                )
            assert (str(e.value.message)) == expected_exc

            # GROUP BY is missing
            previous_group_by = "column_1, column_2, column_3"
            with open(Path(self.path) / "query.sql", "w") as f:
                f.write("SELECT column_1, column_2 FROM upstream_1")
            with open(Path(self.path_previous) / "query.sql", "w") as f:
                f.write(
                    f"SELECT column_1, column_2 FROM upstream_1 GROUP BY {previous_group_by}"
                )
            with pytest.raises(ClickException) as e:
                generate_query_with_shredder_mitigation(
                    client=mock_client,
                    project_id=self.project_id,
                    dataset=self.dataset,
                    destination_table=self.destination_table,
                    staging_table_name=self.staging_table_name,
                    backfill_date=PREVIOUS_DATE,
                )
            assert (str(e.value.message)) == expected_exc

    @patch("google.cloud.bigquery.Client")
    @patch("bigquery_etl.backfill.shredder_mitigation.classify_columns")
    def test_generate_query_called_with_correct_parameters(
        self, mock_classify_columns, mock_client, runner
    ):
        """Test that function generate_query is called with the correct parameters."""
        existing_schema = {
            "fields": [
                {"name": "column_1", "type": "DATE", "mode": "NULLABLE"},
                {"name": "column_2", "type": "STRING", "mode": "NULLABLE"},
                {"name": "metric_1", "type": "INTEGER", "mode": "NULLABLE"},
            ]
        }
        new_schema = {
            "fields": [
                {"name": "column_1", "type": "DATE", "mode": "NULLABLE"},
                {"name": "column_2", "type": "STRING", "mode": "NULLABLE"},
                {"name": "column_3", "type": "STRING", "mode": "NULLABLE"},
                {"name": "metric_1", "type": "INTEGER", "mode": "NULLABLE"},
            ]
        }

        with runner.isolated_filesystem():
            os.makedirs(self.path, exist_ok=True)
            os.makedirs(self.path_previous, exist_ok=True)

            with open(Path(self.path) / "query.sql", "w") as f:
                f.write(
                    "SELECT column_1, column_2, column_3 FROM upstream_1 "
                    "GROUP BY column_1, column_2, column_3"
                )
            with open(Path(self.path) / "metadata.yaml", "w") as f:
                f.write(
                    "bigquery:\n  time_partitioning:\n    type: day\n    "
                    "field: column_1\n    require_partition_filter: true"
                )
            with open(Path(self.path_previous) / "query.sql", "w") as f:
                f.write(
                    "SELECT column_1, column_2 FROM upstream_1 GROUP BY column_1, column_2"
                )
            with open(Path(self.path) / "metadata.yaml", "w") as f:
                f.write(
                    "bigquery:\n  time_partitioning:\n    type: day\n    "
                    "field: column_1\n    require_partition_filter: true\n"
                    "labels:\n    shredder_mitigation: true"
                )
            with open(Path(self.path_previous) / "metadata.yaml", "w") as f:
                f.write(
                    "bigquery:\n  time_partitioning:\n    type: day\n    "
                    "field: column_1\n    require_partition_filter: true\n"
                    "labels:\n    shredder_mitigation: true"
                )
            with open(self.path / "schema.yaml", "w") as f:
                f.write(yaml.safe_dump(new_schema))
            with open(
                self.path_previous / "schema.yaml",
                "w",
            ) as f:
                f.write(yaml.safe_dump(existing_schema))

            mock_classify_columns.return_value = (
                [
                    Column(
                        "column_1",
                        DataTypeGroup.DATE,
                        ColumnType.DIMENSION,
                        ColumnStatus.COMMON,
                    ),
                    Column(
                        "column_2",
                        DataTypeGroup.STRING,
                        ColumnType.DIMENSION,
                        ColumnStatus.COMMON,
                    ),
                ],
                [
                    Column(
                        "column_3",
                        DataTypeGroup.STRING,
                        ColumnType.DIMENSION,
                        ColumnStatus.ADDED,
                    )
                ],
                [],
                [
                    Column(
                        "metric_1",
                        DataTypeGroup.INTEGER,
                        ColumnType.METRIC,
                        ColumnStatus.COMMON,
                    )
                ],
                [],
            )

            with patch.object(
                Subset,
                "get_query_path_results",
                return_value=[{"column_1": "ABC", "column_2": "DEF", "metric_1": 10.0}],
            ):
                with patch.object(Subset, "generate_query") as mock_generate_query:
                    generate_query_with_shredder_mitigation(
                        client=mock_client,
                        project_id=self.project_id,
                        dataset=self.dataset,
                        destination_table=self.destination_table,
                        staging_table_name=self.staging_table_name,
                        backfill_date=PREVIOUS_DATE,
                    )
                    assert mock_generate_query.call_count == 5
                    assert mock_generate_query.call_args_list == (
                        [
                            call(
                                select_list=[
                                    "column_1",
                                    "IF(column_2 IS NULL OR column_2 = '??', '???????', column_2) AS column_2",
                                    "SUM(metric_1) AS metric_1",
                                ],
                                from_clause="new_version",
                                group_by_clause="ALL",
                            ),
                            call(
                                select_list=[
                                    "column_1",
                                    "IF(column_2 IS NULL OR column_2 = '??', '???????', column_2) AS column_2",
                                    "SUM(metric_1) AS metric_1",
                                ],
                                from_clause="`moz-fx-data-shared-prod.test.test_query_v1`",
                                where_clause="column_1 = @column_1",
                                group_by_clause="ALL",
                            ),
                            call(
                                select_list=[
                                    "previous_agg.column_1",
                                    "previous_agg.column_2",
                                    "CAST(NULL AS STRING) AS column_3",
                                    "COALESCE(previous_agg.metric_1, 0) - "
                                    "COALESCE(new_agg.metric_1, 0) AS metric_1",
                                ],
                                from_clause="previous_agg LEFT JOIN new_agg ON "
                                "previous_agg.column_1 = new_agg.column_1"
                                " AND previous_agg.column_2 = new_agg.column_2 ",
                                where_clause="COALESCE(previous_agg.metric_1, 0) >"
                                " COALESCE(new_agg.metric_1, 0)",
                            ),
                            call(
                                select_list=[
                                    "column_1",
                                    "IF(column_2 IS NULL OR column_2 = '??', '???????', column_2) AS column_2",
                                    "SUM(metric_1) AS metric_1",
                                ],
                                from_clause="`moz-fx-data-shared-prod.test.test_query_v1`",
                                where_clause="column_1 = @column_1",
                                group_by_clause="ALL",
                            ),
                            call(
                                select_list=[
                                    "column_1",
                                    "IF(column_2 IS NULL OR column_2 = '??', '???????', column_2) AS column_2",
                                    "SUM(metric_1) AS metric_1",
                                ],
                                from_clause="`moz-fx-data-shared-prod.test.test_query_v2__2021_01_01`",
                                where_clause="column_1 = @column_1",
                                group_by_clause="ALL",
                            ),
                        ]
                    )

    @patch("google.cloud.bigquery.Client")
    @patch("bigquery_etl.backfill.shredder_mitigation.classify_columns")
    def test_generate_query_and_shredder_mitigation_label(
        self, mock_classify_columns, mock_client, runner, capfd
    ):
        """Test that query is generated as expected given a set of mock dimensions and metrics."""
        expected_failure_output = (
            "The required label `shredder_mitigation` is missing in the metadata of the "
            "table. The process will now terminate.\n"
        )
        existing_schema = {
            "fields": [
                {"name": "column_1", "type": "DATE", "mode": "NULLABLE"},
                {"name": "metric_1", "type": "INTEGER", "mode": "NULLABLE"},
            ]
        }
        new_schema = {
            "fields": [
                {"name": "column_1", "type": "DATE", "mode": "NULLABLE"},
                {"name": "column_2", "type": "STRING", "mode": "NULLABLE"},
                {"name": "metric_1", "type": "INTEGER", "mode": "NULLABLE"},
            ]
        }

        with runner.isolated_filesystem():
            os.makedirs(self.path, exist_ok=True)
            os.makedirs(self.path_previous, exist_ok=True)

            with open(Path(self.path) / "query.sql", "w") as f:
                f.write("SELECT column_1 FROM upstream_1 GROUP BY column_1")
            with open(Path(self.path_previous) / "query.sql", "w") as f:
                f.write(
                    "SELECT column_1, column_2 FROM upstream_1 GROUP BY column_1, column_2"
                )

            with open(self.path / "schema.yaml", "w") as f:
                f.write(yaml.safe_dump(new_schema))
            with open(self.path_previous / "schema.yaml", "w") as f:
                f.write(yaml.safe_dump(existing_schema))

            # Label missing in both versions.
            with open(Path(self.path) / "metadata.yaml", "w") as f:
                f.write("Friendly name: Test\n")
            with open(Path(self.path_previous) / "metadata.yaml", "w") as f:
                f.write("Friendly name: Test\n")

            with pytest.raises(SystemExit) as result:
                generate_query_with_shredder_mitigation(
                    client=mock_client,
                    project_id=self.project_id,
                    dataset=self.dataset,
                    destination_table=self.destination_table,
                    staging_table_name=self.staging_table_name,
                    backfill_date=PREVIOUS_DATE,
                )
            assert result.type == SystemExit
            assert result.value.code == 1
            captured = capfd.readouterr()
            assert expected_failure_output == captured.out

            # Label missing in backfilled version generates failure even if in previous version.
            with open(Path(self.path_previous) / "metadata.yaml", "w") as f:
                f.write("Friendly name: Test\nlabels:\n    shredder_mitigation: true")
            with open(Path(self.path) / "metadata.yaml", "w") as f:
                f.write("Friendly name: Test\n")

            with pytest.raises(SystemExit) as result:
                generate_query_with_shredder_mitigation(
                    client=mock_client,
                    project_id=self.project_id,
                    dataset=self.dataset,
                    destination_table=self.destination_table,
                    staging_table_name=self.staging_table_name,
                    backfill_date=PREVIOUS_DATE,
                )
            assert result.type == SystemExit
            assert result.value.code == 1
            captured = capfd.readouterr()
            assert expected_failure_output == captured.out

            # Label missing in previous version doesn't generate failure
            # as long as it's present in backfilled version.
            with open(Path(self.path_previous) / "metadata.yaml", "w") as f:
                f.write("Friendly name: Test\n")
            with open(Path(self.path) / "metadata.yaml", "w") as f:
                f.write(
                    "Friendly name: Test\n" "labels:\n    shredder_mitigation: true"
                )

            mock_classify_columns.return_value = (
                [
                    Column(
                        "column_1",
                        DataTypeGroup.DATE,
                        ColumnType.DIMENSION,
                        ColumnStatus.COMMON,
                    )
                ],
                [
                    Column(
                        "column_2",
                        DataTypeGroup.DATE,
                        ColumnType.DIMENSION,
                        ColumnStatus.COMMON,
                    )
                ],
                [],
                [
                    Column(
                        "metric_1",
                        DataTypeGroup.INTEGER,
                        ColumnType.METRIC,
                        ColumnStatus.COMMON,
                    )
                ],
                [],
            )

            with patch.object(
                Subset,
                "get_query_path_results",
                return_value=[{"column_1": "ABC", "column_2": "DEF", "metric_1": 10.0}],
            ):
                result = generate_query_with_shredder_mitigation(
                    client=mock_client,
                    project_id=self.project_id,
                    dataset=self.dataset,
                    destination_table=self.destination_table,
                    staging_table_name=self.staging_table_name,
                    backfill_date=PREVIOUS_DATE,
                )
                assert result[0] == self.path



from argparse import ArgumentParser
import requests
from requests.auth import HTTPBasicAuth
import json
import os
import sys

from google.cloud import bigquery
import google.auth
import logging


class BigQueryAPI:
    def __init__(self) -> None:
        self.logger = logging.getLogger(self.__class__.__name__)

    def load_jira_user_data(self, destination_table: str, users: list[dict], truncate=True):
        """Load downloaded data to BQ table.
        https://cloud.google.com/bigquery/docs/loading-data-cloud-storage-json#python
        """
        credentials, project = google.auth.default(
            scopes=[
                "https://www.googleapis.com/auth/cloud-platform",
                "https://www.googleapis.com/auth/bigquery",
            ]
        )
        client = bigquery.Client(credentials=credentials, project=project)
        write_disposition = bigquery.WriteDisposition.WRITE_TRUNCATE if truncate else bigquery.WriteDisposition.WRITE_APPEND
        
        job_config = bigquery.LoadJobConfig(
            schema=[
                bigquery.SchemaField("account_id", "STRING"),
                bigquery.SchemaField("account_status", "STRING"),                
                bigquery.SchemaField("name", "STRING"),
                bigquery.SchemaField("time_zone", "STRING"),
                bigquery.SchemaField("email_address", "STRING"),
                bigquery.SchemaField("account_type", "STRING"),
                bigquery.SchemaField("locale", "STRING"),
            ],
            autodetect=False,
            write_disposition=write_disposition,
            source_format=bigquery.SourceFormat.NEWLINE_DELIMITED_JSON,
        )

        job = client.load_table_from_json(
            users, destination_table, job_config=job_config
        )
        job.result()



class JiraAPI:
    def __init__(self, args) -> None:
        self.logger = logging.getLogger(self.__class__.__name__)

        self.secrets_dict = {
          "jira_username": "$JIRA_USERNAME",
          "jira_token": "$JIRA_TOKEN",
        }
        self.secrets_dict = {key:os.path.expandvars(self.secrets_dict[key]) for (key,value) in self.secrets_dict.items() }
        self.base_jira_url = args.base_jira_url
        self.auth = HTTPBasicAuth(self.secrets_dict.get('jira_username'), self.secrets_dict.get('jira_token'))

    def get_users_paged(self, max_results=500):
        startAt = 0
        headers = {"Accept": "application/json"}

        while True:           
            url = (
                self.base_jira_url
                + f"/rest/api/3/users/search?query=+&maxResults={max_results}&startAt={startAt}"
            )
            try:
                response = requests.request("GET", url, headers=headers, auth=self.auth)
            except Exception as e:
                self.logger.error(str(e))
                self.logger.critical("Failed while getting Jira users")
                sys.exit(1)
            is_success = 299 >= response.status_code >= 200
            if not is_success:
                self.logger.error(f"ERROR: response.status_code = {response.status_code}")
                self.logger.error(f"ERROR: response.text = {response.text}")
                self.logger.error(f"ERROR: response.reason = {response.reason}")
                self.logger.critical("Failed while getting Jira users")
                sys.exit(1)
                
            users = response.json()
            yield [
                {
                    "account_id": user.get("accountId", ""),
                    "account_status": (
                        "active" if user.get("active", "") else "inactive"
                    ),                    
                    "name": user.get("displayName", ""),
                    "time_zone": user.get("timeZone", ""),                    
                    "email_address": user.get("emailAddress", ""),
                    "account_type": user.get("accountType", ""),
                    "locale": user.get("locale", ""),                    
                }
                for user in users
            ] 
            
            if len(users)<max_results:
                break
            
            startAt +=max_results    
             
             
class JiraBigQueryIntegration:

    def __init__(self) -> None:
        self.logger = logging.getLogger(self.__class__.__name__)

    def run(self, args):
        
        self.logger.info("Starting Jira BigQuery Integration ...")
        jira = JiraAPI(args)
        bigquery = BigQueryAPI()
        
        truncate = True          
        for users in jira.get_users_paged():                       
            bigquery.load_jira_user_data(args.destination, users,truncate)
            self.logger.info(f"Added {len(users)} users to user table")
            truncate = False    
        
        self.logger.info("End of Jira BigQuery Integration")
        
        
def main():
    parser = ArgumentParser()
    parser.add_argument(
        "--destination",
        dest="destination",
        default="moz-fx-data-shared-prod.jira_service_desk_derived.user_v1",
        required=False,
    )
    parser.add_argument(
        "--base-url",
        dest="base_jira_url",
        default="https://mozilla-hub-sandbox-721.atlassian.net",
        required=False
    )

    args = parser.parse_args()

    logging.basicConfig(
        format="%(asctime)s:\t%(name)s.%(funcName)s()[%(filename)s:%(lineno)s]:\t%(levelname)s: %(message)s",
        level=logging._checkLevel("INFO"),
        encoding="utf-8",
    )

    integration = JiraBigQueryIntegration()
    integration.run(args)


main()



#!/usr/bin/env python3
"""
Script for initializing tests for queries related to glam-fenix-dev.

This only needs to be run once for the initial testing.
"""
import json
import shutil
import warnings
from collections import namedtuple
from glob import glob
from multiprocessing import Pool
from pathlib import Path

import click
import yaml
from google.cloud import bigquery

from bigquery_etl.glam.utils import run

warnings.filterwarnings(
    "ignore", "Your application has authenticated using end user credentials"
)

ROOT = Path(run("git rev-parse --show-toplevel"))
SQL_ROOT = ROOT / "sql" / "glam-fenix-dev" / "glam_etl"
TEST_ROOT = Path(__file__).parent


def list_queries(path):
    """List all of the queries in the glam_etl folder."""
    return list([p for p in path.glob("*") if "__clients_daily" not in p.name])


def dryrun(sql):
    """Dry run a query and return the referenced tables.

    This does not capture views, but only the tables referenced by those
    views.
    """
    client = bigquery.Client(project="glam-fenix-dev")
    job = client.query(
        sql,
        job_config=bigquery.QueryJobConfig(
            dry_run=True,
            use_query_cache=False,
            query_parameters=[
                bigquery.ScalarQueryParameter("submission_date", "DATE", "2020-10-01"),
                bigquery.ScalarQueryParameter("min_sample_id", "INT64", 0),
                bigquery.ScalarQueryParameter("max_sample_id", "INT64", 99),
                bigquery.ScalarQueryParameter("sample_size", "INT64", 100),
            ],
        ),
    )
    return job.referenced_tables


def _dryrun(query_path):
    """Return the dry run results from a path."""
    return (query_path, dryrun(query_path.read_text()))


def calculate_dependencies(queries):
    """Create a list of dependencies between queries for bootstrapping schemas."""
    with Pool(20) as p:
        dryrun_references = p.map(_dryrun, queries)
    res = []
    for query, refs in dryrun_references:
        for ref in refs:
            res.append(
                (
                    f"{ref.project}:{ref.dataset_id}.{ref.table_id}",
                    f"{query.parent.parent.parent.name}:"
                    f"{query.parent.parent.name}.{query.parent.name}",
                )
            )
    return sorted(set(res))


@click.group()
def bootstrap():
    """Script for initializing tests for queries related to glam-fenix-dev."""
    pass


@bootstrap.command()
@click.option("--output", default=TEST_ROOT / "dependencies.json")
def deps(output):
    """Create a dependency file with all links between queries and tables."""
    path = Path(output)
    deps = calculate_dependencies(
        [
            p
            for p in map(Path, glob(f"{SQL_ROOT}/**/*.sql", recursive=True))
            if "__clients_daily" not in p.name
        ]
    )
    path.write_text(
        json.dumps([dict(zip(["from", "to"], dep)) for dep in deps], indent=2)
    )


@bootstrap.command()
@click.option(
    "--dependency-file",
    default=TEST_ROOT / "dependencies.json",
    type=click.Path(dir_okay=False, exists=True),
)
def mermaid(dependency_file):
    """Generate a mermaid diagram of dependencies."""
    deps = json.loads(Path(dependency_file).read_text())

    # project, dataset, product, query
    Node = namedtuple("node", ["qualified", "project", "dataset", "name", "product"])

    def parse(x):
        project, dataset_table = x.split(":")
        dataset, table = dataset_table.split(".")
        is_glam = dataset == "glam_etl"
        return Node(
            dataset_table if is_glam else x,
            project,
            dataset,
            table.split("__")[1] if is_glam else table,
            table.split("__")[0] if is_glam else None,
        )

    # build up mapping of nodes to parsed info
    nodes = {}
    for dep in deps:
        src = dep["from"]
        dst = dep["to"]
        for x in [src, dst]:
            if "*" not in x:
                nodes[x] = parse(x)

    print("graph LR")
    projects = set(n.project for n in nodes.values())
    for project in projects:
        print(f'subgraph "{project}"')
        datasets = set(n.dataset for n in nodes.values() if n.project == project)
        for dataset in datasets:
            print(f'subgraph "{dataset}"')
            if dataset == "glam_etl":
                products = set(
                    n.product for n in nodes.values() if n.dataset == "glam_etl"
                )
                for product in products:
                    print(f'subgraph "{product}"')
                    for n in set(
                        n
                        for n in nodes.values()
                        if n.dataset == "glam_etl" and n.product == product
                    ):
                        print(f"{n.qualified}[{n.name}]")
                    print("end")
            else:
                for n in set(n for n in nodes.values() if n.dataset == dataset):
                    print(f"{n.qualified}[{n.name}]")
            print("end")
        print("end")

    for dep in deps:
        src = dep["from"]
        dst = dep["to"]
        if "*" in src:
            for node in [
                v for k, v in nodes.items() if k.startswith(src.replace("*", ""))
            ]:
                print(f"{node.qualified} --> {nodes[dst].qualified}")
        else:
            print(f"{nodes[src].qualified} --> {nodes[dst].qualified}")


@bootstrap.command()
@click.option(
    "--dependency-file",
    default=TEST_ROOT / "dependencies.json",
    type=click.Path(dir_okay=False, exists=True),
)
def skeleton(dependency_file):
    """Generate the skeleton for minimal tests.

    Schemas are populated from the sql folder. Run the `deps` command to
    generate a file of dependencies between queries.
    """
    deps = json.loads(Path(dependency_file).read_text())
    for query in list_queries(SQL_ROOT):
        # only tests for org_mozilla_fenix_glam_nightly
        if "org_mozilla_fenix_glam_nightly" not in query.name:
            continue

        print(f"copying schemas for {query.name}")
        # views are also considered aggregates
        path = TEST_ROOT / query.name / "test_minimal"
        path.mkdir(parents=True, exist_ok=True)

        # set query parameters
        params = list(
            map(
                lambda x: dict(zip(["name", "type", "value"], x)),
                [
                    ("submission_date", "DATE", "2020-10-01"),
                    ("min_sample_id", "INT64", 0),
                    ("max_sample_id", "INT64", 99),
                    ("sample_size", "INT64", 100),
                ],
            )
        )
        with (path / "query_params.yaml").open("w") as fp:
            yaml.dump(params, fp)

        # now copy over the schemas
        query_deps = [x["from"] for x in deps if query.name in x["to"]]
        print(f"found dependencies: {json.dumps(query_deps, indent=2)}")
        for dep in query_deps:
            project = dep.split(":")[0]
            dataset = dep.split(":")[1].split(".")[0]
            table = dep.split(".")[1]
            for schema in (ROOT / "sql" / project / dataset).glob(f"{table}/schema.*"):
                print(f"copied dependency {schema}")
                shutil.copyfile(
                    schema,
                    path
                    / (
                        (
                            f"{project}.{dataset}."
                            if dataset != "glam_etl"
                            else f"{dataset}."
                        )
                        + f"{schema.parent.name}.{schema.name}"
                    ),
                )


if __name__ == "__main__":
    bootstrap()



"""Standard definitions for reusable script arguments."""

import fnmatch
import logging
import re
from argparse import Action
from functools import partial

from google.cloud import bigquery

from bigquery_etl.config import ConfigLoader
from bigquery_etl.util.common import TempDatasetReference


def add_argument(parser, *args, **kwargs):
    """Add default to help while adding argument to parser."""
    if "help" in kwargs:
        default = kwargs.get("default")
        if default not in (None, [], [None]):
            if kwargs.get("nargs") in ("*", "+"):
                # unnest a single default for printing, if possible
                try:
                    (default,) = default
                except ValueError:
                    pass
            kwargs["help"] += f"; Defaults to {default}"
    parser.add_argument(*args, **kwargs)


def add_billing_projects(parser, *extra_args, default=[None]):
    """Add argument for billing projects."""
    add_argument(
        parser,
        "-p",
        "--billing-projects",
        "--billing_projects",
        "--billing-project",
        "--billing_project",
        *extra_args,
        nargs="+",
        default=default,
        help="One or more billing projects over which bigquery jobs should be "
        "distributed",
    )


def add_dry_run(parser, debug_log_queries=True):
    """Add argument for dry run."""
    add_argument(
        parser,
        "--dry_run",
        "--dry-run",
        action="store_true",
        help="Do not make changes, only log actions that would be taken"
        + (
            "; Use with --log-level=DEBUG to log query contents"
            if debug_log_queries
            else ""
        ),
    )


def add_log_level(parser, default=logging.getLevelName(logging.INFO)):
    """Add argument for log level."""
    add_argument(
        parser,
        "-l",
        "--log-level",
        "--log_level",
        action=LogLevelAction,
        default=default,
        type=str.upper,
        help="Set logging level for the python root logger",
    )


def add_parallelism(parser, default=4):
    """Add argument for parallel execution."""
    add_argument(
        parser,
        "-P",
        "--parallelism",
        default=default,
        type=int,
        help="Maximum number of tasks to execute concurrently",
    )


def add_priority(parser):
    """Add argument for BigQuery job priority."""
    add_argument(
        parser,
        "--priority",
        default=bigquery.QueryPriority.INTERACTIVE,
        type=str.upper,
        choices=[bigquery.QueryPriority.BATCH, bigquery.QueryPriority.INTERACTIVE],
        help="Priority for BigQuery query jobs; BATCH priority may significantly slow "
        "down queries if reserved slots are not enabled for the billing project; "
        "INTERACTIVE priority is limited to 100 concurrent queries per project",
    )


def add_table_filter(parser, example="telemetry_stable.main_v*"):
    """Add arguments for filtering tables."""
    example_ = f"Pass names or globs like {example!r}"
    add_argument(
        parser,
        "-o",
        "--only",
        nargs="+",
        dest="table_filter",
        raw_dest="only_tables",
        action=TableFilterAction,
        help=f"Process only the given tables; {example_}",
    )
    add_argument(
        parser,
        "-x",
        "--except",
        nargs="+",
        dest="table_filter",
        raw_dest="except_tables",
        action=TableFilterAction,
        help=f"Process all tables except for the given tables; {example_}",
    )


def add_temp_dataset(parser, *extra_args):
    """Add argument for temporary dataset."""
    add_argument(
        parser,
        "--temp-dataset",
        "--temp_dataset",
        "--temporary-dataset",
        "--temporary_dataset",
        *extra_args,
        default=f"{ConfigLoader.get('default', 'project', fallback='moz-fx-data-shared-prod')}.tmp",
        type=TempDatasetReference.from_string,
        help="Dataset where intermediate query results will be temporarily stored, "
        "formatted as PROJECT_ID.DATASET_ID",
    )


class LogLevelAction(Action):
    """Custom argparse.Action for --log-level."""

    def __init__(self, *args, **kwargs):
        """Set default log level if provided."""
        super().__init__(*args, **kwargs)
        if self.default is not None:
            logging.root.setLevel(self.default)

    def __call__(self, parser, namespace, value, option_string=None):
        """Set level for root logger."""
        logging.root.setLevel(value)


class TableFilterAction(Action):
    """Custom argparse.Action for --only and --except."""

    def __init__(self, *args, raw_dest, **kwargs):
        """Add default."""
        super().__init__(*args, default=self.default, **kwargs)
        self.raw_dest = raw_dest
        self.arg = self.option_strings[-1]
        self.invert_match = self.arg == "--except"

    @staticmethod
    def default(table):
        """Return True for default predicate."""
        return True

    @staticmethod
    def compile(values):
        """Compile a list of glob patterns into a single regex."""
        return re.compile("|".join(fnmatch.translate(pattern) for pattern in values))

    def predicate(self, table, pattern):
        """Log tables skipped due to table filter arguments."""
        matched = (pattern.match(table) is not None) != self.invert_match
        if not matched:
            logging.info(f"Skipping {table} due to {self.arg} argument")
        return matched

    def __call__(self, parser, namespace, values, option_string=None):
        """Add table filter to predicates."""
        setattr(namespace, self.raw_dest, values)
        predicates_attr = "_" + self.dest
        predicates = getattr(namespace, predicates_attr, [])
        if not hasattr(namespace, predicates_attr):
            setattr(namespace, predicates_attr, predicates)
            setattr(
                namespace,
                self.dest,
                lambda table: all(predicate(table) for predicate in predicates),
            )
        predicates.append(partial(self.predicate, pattern=self.compile(values)))



"""Import data from alchemer (surveygizmo) surveys into BigQuery."""

import datetime as dt
import json
import re
from pathlib import Path

import click
import pytz
import requests
from google.cloud import bigquery
from requests import HTTPError


def utc_date_to_eastern_string(date_string):
    """Normalize ISO date to UTC midnight."""
    naive_dt = dt.datetime.strptime(date_string, "%Y-%m-%d")
    as_utc = pytz.utc.localize(naive_dt)
    as_eastern = as_utc.astimezone(pytz.timezone("US/Eastern"))
    return dt.datetime.strftime(as_eastern, "%Y-%m-%d+%H:%M:%S")


def date_plus_one(date_string):
    """Get the day after the current date."""
    return dt.datetime.strftime(
        dt.date.fromisoformat(date_string) + dt.timedelta(days=1), "%Y-%m-%d"
    )


def format_responses(s, date):
    """Return a nested field of responses for each user."""
    # `survey_data` is a dict with question ID as the key and question/response
    # details as the value e.g. survey_data: {'25': {'id': 25, 'type': 'RADIO',
    # 'question': 'I trust Firefox to help me with my online privacy',
    # 'section_id': 2, 'answer': 'Agree', 'answer_id': 10066, 'shown': True}}
    # See https://apihelp.alchemer.com/help/surveyresponse-returned-fields-v5#getobject

    # Note that we are omitted date_submission and date_completed because the
    # timezone is not ISO compliant. The submission date that is passed in as
    # the time parameter should suffice.
    fields = ["id", "session_id", "status", "response_time"]
    results = []
    for data in s.get("survey_data", {}).values():
        # There can be answer_id's like "123456-other"
        if data.get("answer_id") and isinstance(data["answer_id"], str):
            numeric = re.findall(r"\d+", data["answer_id"])
            if numeric:
                data["answer_id"] = int(numeric[0])
            else:
                del data["answer_id"]
        if data.get("options"):
            data["options"] = list(data["options"].values())
        if data.get("subquestions"):
            data["subquestions"] = json.dumps(data["subquestions"])
        results.append(data)

    return {
        # this is used as the partitioning field
        "submission_date": date,
        **{field: s[field] for field in fields},
        "survey_data": results,
    }


def construct_data(survey, date):
    """Construct response data."""
    return [format_responses(resp, date) for resp in survey["data"]]


def get_survey_data(survey_id, date_string, token, secret):
    """Get survey data from a survey id and date."""
    # per SurveyGizmo docs, times are assumed to be eastern
    # https://apihelp.surveygizmo.com/help/filters-v5
    # so UTC midnight must be converted to EST/EDT
    start_date = utc_date_to_eastern_string(date_string)
    end_date = utc_date_to_eastern_string(date_plus_one(date_string))

    url = (
        f"https://restapi.surveygizmo.com/v5/survey/{survey_id}/surveyresponse"
        f"?api_token={token}&api_token_secret={secret}&results_per_page=500"
        # filter for date_submitted >= start_date
        f"&filter[field][0]=date_submitted"
        f"&filter[operator][0]=>="
        f"&filter[value][0]={start_date}"
        # filter for date_submitted < end_date
        f"&filter[field][1]=date_submitted"
        f"&filter[operator][1]=<"
        f"&filter[value][1]={end_date}"
    )
    resp = _get_request(url)
    survey = resp.json()

    # if the result set is large, we'll have to page through them to get all data
    total_pages = survey.get("total_pages")
    print(f"Found {total_pages} pages after filtering on date={date_string}")

    print("fetching page 1")
    ret = construct_data(survey, date_string)

    for page in range(2, total_pages + 1):
        print(f"fetching page {page}")
        resp = _get_request(url + f"&page={page}")
        ret = ret + construct_data(resp.json(), date_string)
    return ret


def _get_request(url):
    """Make get request and print response if there is an exception."""
    resp = requests.get(url)
    try:
        resp.raise_for_status()
    except HTTPError:
        print(f"Error response: {resp.text}")
        raise
    return resp


def response_schema():
    """Get the schema for the response object from disk."""
    path = Path(__file__).parent / "response.schema.json"
    return bigquery.SchemaField.from_api_repr(
        {"name": "root", "type": "RECORD", "fields": json.loads(path.read_text())}
    ).fields


def insert_to_bq(
    data, table, date, write_disposition=bigquery.job.WriteDisposition.WRITE_TRUNCATE
):
    """Insert data into a bigquery table."""
    client = bigquery.Client()
    print(f"Inserting {len(data)} rows into bigquery")
    job_config = bigquery.LoadJobConfig(
        # We may also infer the schema by setting `autodetect=True`
        schema=response_schema(),
        schema_update_options=bigquery.SchemaUpdateOption.ALLOW_FIELD_ADDITION,
        write_disposition=write_disposition,
        time_partitioning=bigquery.table.TimePartitioning(field="submission_date"),
    )
    partition = f"{table}${date.replace('-', '')}"
    job = client.load_table_from_json(data, partition, job_config=job_config)
    print(f"Running job {job.job_id}")
    # job.result() returns a LoadJob object if successful, or raises an exception if not
    job.result()


@click.command()
@click.option("--date", required=True)
@click.option("--survey_id", required=True)
@click.option("--api_token", required=True)
@click.option("--api_secret", required=True)
@click.option("--destination_table", required=True)
def main(date, survey_id, api_token, api_secret, destination_table):
    """Import data from alchemer (surveygizmo) surveys into BigQuery."""
    survey_data = get_survey_data(survey_id, date, api_token, api_secret)
    insert_to_bq(survey_data, destination_table, date)


if __name__ == "__main__":
    main()



"""
Validate HMAC-SHA256 implementation against the NIST test vectors.

The vectors are located in tests/validation/data/hmac_sha256_validation.json.
"""

import json

import pytest
from google.cloud import bigquery

from bigquery_etl.routine.parse_routine import (
    RawRoutine,
    read_routine_dir,
    routine_tests_sql,
)

validation_data_file = "tests/validation/data/hmac_sha256_validation.json"


def udfs():
    """Get all udfs and assertions."""
    return read_routine_dir("sql/mozfun/assert", "udf", "udf_js")


def load_data():
    """Load test data."""
    with open(validation_data_file, "r") as f:
        return json.load(f)["data"]


def generate_raw_routine(test_cases):
    """Generate a SQL test for each instance in hmac_sha256_validation.json."""
    test_sql_fixture = (
        "SELECT assert_equals("
        "'{Mac}',"
        "TO_HEX(SUBSTR("
        "udf.hmac_sha256("
        "FROM_HEX('{Key}'),"
        "FROM_HEX('{Msg}')),"
        "1,"
        "{Tlen})));"
    )
    test_sql_stmnts = [test_sql_fixture.format(**test_case) for test_case in test_cases]

    return RawRoutine.from_file(
        from_text="\n".join(test_sql_stmnts),
        path="udf" / "hmac_sha256" / "hmac_sha256" / "sql",
    )


def generate_sql():
    """Generate SQL statements to test."""
    return routine_tests_sql(generate_raw_routine(load_data()), udfs())


@pytest.mark.parametrize("sql", generate_sql())
def test_validate_hmac_sha256(sql):
    """Validate hmac_sha256."""
    job_config = bigquery.QueryJobConfig(use_legacy_sql=False)
    job = bigquery.Client().query(sql, job_config=job_config)
    job.result()



"""PyTest configuration."""

import os
import random
import string
import subprocess
from pathlib import Path

import pytest
from google.cloud import bigquery, storage

TEST_BUCKET = "bigquery-etl-integration-test-bucket"


pytest_plugins = [
    "bigquery_etl.pytest_plugin.sql",
    "bigquery_etl.pytest_plugin.routine",
    "bigquery_etl.pytest_plugin.script_lint.black",
    "bigquery_etl.pytest_plugin.script_lint.docstyle",
    "bigquery_etl.pytest_plugin.script_lint.flake8",
    "bigquery_etl.pytest_plugin.script_lint.mypy",
]


def pytest_collection_modifyitems(config, items):
    keywordexpr = config.option.keyword
    markexpr = config.option.markexpr
    if keywordexpr or markexpr:
        return

    skip_integration = pytest.mark.skip(reason="integration marker not selected")

    for item in items:
        if "integration" in item.keywords:
            item.add_marker(skip_integration)


@pytest.fixture
def project_id():
    """Provide a BigQuery project ID."""
    # GOOGLE_PROJECT_ID needs to be set for integration tests to run
    project_id = os.environ["GOOGLE_PROJECT_ID"]

    return project_id


@pytest.fixture
def bigquery_client():
    """Provide a BigQuery client."""
    project_id = os.environ["GOOGLE_PROJECT_ID"]
    return bigquery.Client(project_id)


@pytest.fixture
def temporary_dataset():
    """Fixture for creating a random temporary BigQuery dataset."""
    # generate a random test dataset to avoid conflicts when running tests in parallel
    test_dataset = "test_" + "".join(
        random.choice(string.ascii_lowercase) for i in range(12)
    )

    project_id = os.environ["GOOGLE_PROJECT_ID"]
    client = bigquery.Client(project_id)
    client.create_dataset(test_dataset)

    yield test_dataset

    # cleanup and remove temporary dataset
    client.delete_dataset(test_dataset, delete_contents=True, not_found_ok=True)


@pytest.fixture
def test_bucket():
    """Provide a test bucket instance."""
    storage_client = storage.Client()
    bucket = storage_client.bucket(TEST_BUCKET)

    yield bucket


@pytest.fixture
def temporary_gcs_folder():
    """Provide a temporary folder in the GCS test bucket."""
    test_folder = (
        "test_"
        + "".join(random.choice(string.ascii_lowercase) for i in range(12))
        + "/"
    )

    yield test_folder

    storage_client = storage.Client()
    bucket = storage_client.bucket(TEST_BUCKET)

    # delete test folder
    blobs = bucket.list_blobs(prefix=test_folder)

    for blob in blobs:
        blob.delete()


@pytest.fixture
def storage_client():
    """Provide a client instance for cloud storage."""
    yield storage.Client()



"""Generates Airflow DAGs for scheduled queries."""

import copy
import logging
import os
from argparse import ArgumentParser
from pathlib import Path

from bigquery_etl.query_scheduling.dag_collection import DagCollection
from bigquery_etl.query_scheduling.task import Task, UnscheduledTask
from bigquery_etl.util import standard_args
from bigquery_etl.util.common import project_dirs

DEFAULT_DAGS_FILE = "dags.yaml"
QUERY_FILE = "query.sql"
QUERY_PART_FILE = "part1.sql"
SCRIPT_FILE = "script.sql"
PYTHON_SCRIPT_FILE = "query.py"
DEFAULT_DAGS_DIR = "dags"
CHECKS_FILE = "checks.sql"
BIGEYE_FILE = "bigconfig.yml"

parser = ArgumentParser(description=__doc__)
parser.add_argument(
    "--dags_config",
    "--dags-config",
    help="File with DAGs configuration",
    default=DEFAULT_DAGS_FILE,
)
parser.add_argument(
    "--output_dir",
    "--output-dir",
    help="Generated DAGs are written to this output directory.",
    default=DEFAULT_DAGS_DIR,
)
parser.add_argument(
    "--dag_id",
    "--dag_id",
    help="The DAG to generate. Generates all DAGs by default",
    default=None,
)
parser.add_argument(
    "--project_id",
    "--project-id",
    help="Project to generate DAGs for. If not specified, generate for all projects.",
    default=None,
    required=False,
)

standard_args.add_log_level(parser)


def get_dags(project_id, dags_config, sql_dir=None):
    """Return all configured DAGs including associated tasks."""
    tasks = []
    dag_collection = DagCollection.from_file(dags_config)
    for project_dir in project_dirs(project_id, sql_dir=sql_dir):
        # parse metadata.yaml to retrieve scheduling information
        if os.path.isdir(project_dir):
            for root, dirs, files in os.walk(project_dir, followlinks=True):
                try:
                    if QUERY_FILE in files:
                        query_file = os.path.join(root, QUERY_FILE)
                        task = Task.of_query(query_file, dag_collection=dag_collection)
                    elif QUERY_PART_FILE in files:
                        # multipart query
                        query_file = os.path.join(root, QUERY_PART_FILE)
                        task = Task.of_multipart_query(
                            query_file, dag_collection=dag_collection
                        )
                    elif SCRIPT_FILE in files:
                        query_file = os.path.join(root, SCRIPT_FILE)
                        task = Task.of_script(query_file, dag_collection=dag_collection)
                    elif PYTHON_SCRIPT_FILE in files:
                        query_file = os.path.join(root, PYTHON_SCRIPT_FILE)
                        task = Task.of_python_script(
                            query_file, dag_collection=dag_collection
                        )
                    else:
                        continue
                except FileNotFoundError:
                    # query has no metadata.yaml file; skip
                    pass
                except UnscheduledTask:
                    # logging.debug(
                    #     f"No scheduling information for {query_file}."
                    # )
                    #
                    # most tasks lack scheduling information for now
                    pass
                except Exception as e:
                    # in the case that there was some other error, report the query
                    # that failed before exiting
                    logging.error(f"Error processing task for query {query_file}")
                    raise e
                else:
                    if BIGEYE_FILE in files:
                        bigeye_file = os.path.join(root, BIGEYE_FILE)
                        bigeye_task = copy.deepcopy(
                            Task.of_bigeye_check(
                                bigeye_file,
                                dag_collection=dag_collection,
                            )
                        )

                        if bigeye_task.monitoring_enabled:
                            tasks.append(bigeye_task)

                    if CHECKS_FILE in files:
                        checks_file = os.path.join(root, CHECKS_FILE)
                        # todo: validate checks file

                        with open(checks_file, "r") as file:
                            file_contents = file.read()

                            if "#fail" in file_contents:
                                checks_task = copy.deepcopy(
                                    Task.of_dq_check(
                                        checks_file,
                                        is_check_fail=True,
                                        dag_collection=dag_collection,
                                    )
                                )
                                tasks.append(checks_task)

                            if "#warn" in file_contents:
                                checks_task = copy.deepcopy(
                                    Task.of_dq_check(
                                        checks_file,
                                        is_check_fail=False,
                                        dag_collection=dag_collection,
                                    )
                                )
                                tasks.append(checks_task)

                    tasks.append(task)
        else:
            logging.error(
                """
                Invalid project_dir: {}, project_dir must be a directory with
                structure <sql>/<project>/<dataset>/<table>/metadata.yaml.
                """.format(
                    project_dir
                )
            )

    return dag_collection.with_tasks(tasks)


def main():
    """Generate Airflow DAGs."""
    args = parser.parse_args()
    dags_output_dir = Path(args.output_dir)

    dags = get_dags(args.project_id, args.dags_config)
    dags.to_airflow_dags(dags_output_dir, args.dag_id)


if __name__ == "__main__":
    main()



"""Generate query directories."""

import os
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional

import click
import yaml
from jinja2 import Environment, FileSystemLoader

from bigquery_etl.cli.utils import is_valid_project, use_cloud_function_option
from bigquery_etl.format_sql.formatter import reformat

TEMPLATED_FILES = {
    "metadata.yaml",
    "query.sql",
    "stored_procedure.sql",
    "udf.sql",
    "view.sql",
    "schema.yaml",
}


ALLOWED_FILES = TEMPLATED_FILES | {"templating.yaml"}


FILE_PATH = os.path.dirname(__file__)
BASE_DIR = Path(FILE_PATH).parent.parent


@dataclass
class Template:
    """A template, to be filled with args and saved as a file."""

    name: str
    env: Environment

    def render(self, write_path, args):
        """Render this template at the specified write_path with the specified args."""
        fpath = write_path / self.name
        print(f"...Generating {str(fpath)}")

        write_path.mkdir(parents=True, exist_ok=True)

        if "header" not in args:
            args["header"] = "Generated by ./bqetl generate events_daily"

        text = self._get_comment_char(fpath.suffix) + args["header"] + "\n\n"
        text += self.env.get_template(self.name).render(**args)

        if fpath.suffix == ".sql":
            text = reformat(text, trailing_newline=True)

        (write_path / self.name).write_text(text)

    def _get_comment_char(self, suffix, append=" "):
        comment_chars = {
            "sql": "--",
            "yaml": "#",
            "yml": "#",
        }

        return comment_chars[suffix[1:]] + append


@dataclass
class TemplatedDir:
    """A directory of templates, which will be rendered per the templating.yaml."""

    name: str
    path: Path
    env: Optional[Environment] = None

    def generate(self, write_path, dataset=None):
        """Render this TemplatedDir to write_path for the specified dataset."""
        args = self.get_args()
        datasets = self.get_datasets(args, dataset)

        for template in self.get_templates():
            for _dataset in datasets:
                template.render(write_path / _dataset / self.name, args[_dataset])

    def get_datasets(self, args, dataset=None) -> List[str]:
        """Get datasets to process."""
        datasets = list(args.keys())
        if dataset is not None:
            datasets = [d for d in datasets if d == dataset]
        if not datasets:
            raise Exception("Nothing to generate, no datasets found for " + self.name)
        return datasets

    def get_templates(self) -> List[Template]:
        """Get the names of the templates to process."""
        env = self.get_environment()
        return [
            Template(f.name, env)
            for f in self.path.glob("*")
            if str(f.name) in TEMPLATED_FILES
        ]

    def get_environment(self) -> Environment:
        """Get the environment."""
        if self.env is None:
            self.env = Environment(loader=FileSystemLoader(str(self.path)))
        return self.env

    def get_args(self) -> dict:
        """Get all arguments for templating, per-dataset."""
        with open(self.path / "templating.yaml", "r") as f:
            return yaml.safe_load(f) or {}


def get_query_dirs(path):
    """Walk a path to get all templated query dirs."""
    for directory, sub_dirs, files in os.walk(path, followlinks=True):
        non_hidden = {f for f in files if not f.startswith(".")}
        if non_hidden and non_hidden.issubset(ALLOWED_FILES):
            dir_path = Path(directory)
            yield TemplatedDir(dir_path.name, dir_path)


@click.command()
@click.option(
    "--target-project",
    "--target_project",
    help="GCP project ID",
    default="moz-fx-data-shared-prod",
    callback=is_valid_project,
)
@click.option(
    "--path",
    help="Where query directories will be searched for",
    default="sql_generators/events_daily/templates",
)
@click.option(
    "--dataset",
    help="The dataset to run this for. "
    "If none selected, runs on all in the configuration yaml file.",
    default=None,
)
@click.option(
    "--output-dir",
    "--output_dir",
    help="Output directory generated SQL is written to",
    type=click.Path(file_okay=False),
    default="sql",
)
@use_cloud_function_option
def generate(target_project, path, dataset, output_dir, use_cloud_function):
    """Generate queries at the path for project."""
    write_path = Path(output_dir) / target_project
    for query_dir in get_query_dirs(path):
        query_dir.generate(write_path, dataset)



"""PyTest plugin for running sql tests."""

import os

import pytest

from bigquery_etl.format_sql.formatter import reformat


def pytest_configure(config):
    """Register a custom marker."""
    config.addinivalue_line("markers", "format_sql: mark format_sql tests.")


def pytest_collect_file(parent, path):
    """Collect format_sql tests."""
    if path.basename == "expect.sql":
        print(path)
        return FormatTest.from_parent(parent, fspath=path)


class FormatTest(pytest.Item, pytest.File):
    """Test a SQL query."""

    def __init__(self, fspath, parent):
        """Initialize."""
        super().__init__(fspath, parent)
        self._nodeid += "::FORMAT_SQL"
        self.add_marker("format_sql")

    def reportinfo(self):
        """Append fspath to report title."""
        return super().reportinfo()[:2] + (self.fspath.relto(os.getcwd()),)

    def _prunetraceback(self, excinfo):
        """Prune traceback to runtest method."""
        traceback = excinfo.traceback
        ntraceback = traceback.cut(path=__file__)
        excinfo.traceback = ntraceback.filter()

    def runtest(self):
        """Run."""
        with open(self.fspath) as fp:
            expect = fp.read()
        try:
            with open(f"{self.fspath.dirname}/input.sql") as fp:
                query = fp.read()
        except FileNotFoundError:
            query = expect
        assert reformat(query) + "\n" == expect



"""Generate and run baseline_clients_first_seen queries for Glean apps."""

from sql_generators.glean_usage.common import GleanTable

TARGET_TABLE_ID = "baseline_clients_first_seen_v1"
PREFIX = "first_seen"


class BaselineClientsFirstSeenTable(GleanTable):
    """Represents generated baseline_clients_first_seen table."""

    def __init__(self):
        """Initialize baseline_clients_first_seen table."""
        GleanTable.__init__(self)
        self.target_table_id = TARGET_TABLE_ID
        self.prefix = PREFIX
        self.custom_render_kwargs = {}

    def generate_per_app_id(
        self,
        project_id,
        baseline_table,
        output_dir=None,
        use_cloud_function=True,
        app_info=[],
        parallelism=8,
        id_token=None,
    ):
        """Generate per-app_id datasets."""
        self.custom_render_kwargs = dict(
            # do not match on org_mozilla_firefoxreality
            fennec_id=any(
                (f"{app_id}_stable" in baseline_table)
                for app_id in [
                    "org_mozilla_firefox",
                    "org_mozilla_fenix_nightly",
                    "org_mozilla_fennec_aurora",
                    "org_mozilla_firefox_beta",
                    "org_mozilla_fenix",
                ]
            )
        )

        GleanTable.generate_per_app_id(
            self,
            project_id,
            baseline_table,
            output_dir=output_dir,
            app_info=app_info,
            parallelism=parallelism,
            id_token=id_token
        )



from pathlib import Path

from bigquery_etl.routine.parse_routine import sub_local_routines

TEST_DIR = Path(__file__).parent.parent


class TestValidateDocs:
    def test_sql_for_dry_run_no_dependencies(self):
        file = (
            TEST_DIR
            / "data"
            / "test_docs"
            / "generated_docs"
            / "test_dataset1"
            / "examples"
            / "example1.sql"
        )

        sql = sub_local_routines(file.read_text(), project="", raw_routines={})
        assert sql == file.read_text()



"""Utility functions used in generating usage queries on top of Glean."""

import glob
import logging
import os
import re
from collections import namedtuple
from functools import cache
from pathlib import Path

import requests
from jinja2 import Environment, FileSystemLoader, TemplateNotFound

from bigquery_etl.config import ConfigLoader
from bigquery_etl.dryrun import DryRun
from bigquery_etl.schema.stable_table_schema import get_stable_table_schemas
from bigquery_etl.util.common import get_table_dir, render, write_sql

APP_LISTINGS_URL = "https://probeinfo.telemetry.mozilla.org/v2/glean/app-listings"
PATH = Path(os.path.dirname(__file__))

# added as the result of baseline checks being added to the template,
# these apps do not receive a baseline ping causing a very basic row_count
# check to fail. For this reason, the checks relying on this ping
# need to be omitted. For more info see: bug-1868848
BQETL_CHECKS_SKIP_APPS = ConfigLoader.get(
    "generate", "glean_usage", "bqetl_checks", "skip_apps", fallback=[]
)
BIGCONFIG_SKIP_APPS = ConfigLoader.get(
    "generate", "glean_usage", "bigconfig", "skip_apps", fallback=[]
)
BIGCONFIG_SKIP_APPS_METRICS = ConfigLoader.get(
    "generate", "glean_usage", "bigconfig", "skip_app_metrics", fallback=[]
)

APPS_WITH_DISTRIBUTION_ID = ("fenix",)

APPS_WITH_PROFILE_GROUP_ID = ("firefox_desktop",)


def write_dataset_metadata(output_dir, full_table_id, derived_dataset_metadata=False):
    """
    Add dataset_metadata.yaml to public facing datasets.

    Does not overwrite existing dataset_metadata.yaml files.
    """
    d = Path(os.path.join(output_dir, *list(full_table_id.split(".")[-2:])))
    d.parent.mkdir(parents=True, exist_ok=True)
    target = d.parent / "dataset_metadata.yaml"

    public_facing = all(
        [postfix not in d.parent.name for postfix in ("_derived", "_stable")]
    )
    if (derived_dataset_metadata or public_facing) and not target.exists():
        env = Environment(loader=FileSystemLoader(PATH / "templates"))
        if derived_dataset_metadata:
            dataset_metadata = env.get_template("derived_dataset_metadata.yaml")
        else:
            dataset_metadata = env.get_template("dataset_metadata.yaml")
        rendered = dataset_metadata.render(
            {
                "friendly_name": " ".join(
                    [p.capitalize() for p in d.parent.name.split("_")]
                ),
                "dataset": d.parent.name,
            }
        )

        logging.info(f"Writing {target}")
        target.write_text(rendered)


def list_tables(project_id, only_tables, table_filter, table_name="baseline_v1"):
    """Return names of all matching baseline tables in shared-prod."""
    prod_baseline_tables = [
        s.stable_table
        for s in get_stable_table_schemas()
        if s.schema_id == "moz://mozilla.org/schemas/glean/ping/1"
        and s.bq_table == table_name
    ]
    prod_datasets_with_baseline = [t.split(".")[0] for t in prod_baseline_tables]
    stable_datasets = prod_datasets_with_baseline
    if only_tables and not _contains_glob(only_tables):
        # skip list calls when only_tables exists and contains no globs
        return [
            f"{project_id}.{t}"
            for t in only_tables
            if table_filter(t) and t in prod_baseline_tables
        ]
    if only_tables and not _contains_glob(
        _extract_dataset_from_glob(t) for t in only_tables
    ):
        stable_datasets = {_extract_dataset_from_glob(t) for t in only_tables}
        stable_datasets = {
            d
            for d in stable_datasets
            if d.endswith("_stable") and d in prod_datasets_with_baseline
        }
    return [
        f"{project_id}.{d}.{table_name}"
        for d in stable_datasets
        if table_filter(f"{d}.{table_name}")
    ]


@cache
def _get_pings_without_metrics():
    def metrics_in_schema(fields):
        for field in fields:
            if field["name"] == "metrics":
                return True
        return False

    return {
        (schema.bq_dataset_family, schema.bq_table_unversioned)
        for schema in get_stable_table_schemas()
        if "glean" in schema.schema_id and not metrics_in_schema(schema.schema)
    }


def ping_has_metrics(dataset_family: str, unversioned_table: str) -> bool:
    """Return true if the given stable table schema has a metrics column at the top-level."""
    return (dataset_family, unversioned_table) not in _get_pings_without_metrics()


def table_names_from_baseline(baseline_table, include_project_id=True):
    """Return a dict with full table IDs for derived tables and views.

    :param baseline_table: stable table ID in project.dataset.table form
    """
    prefix = re.sub(r"_stable\..+", "", baseline_table)
    if not include_project_id:
        prefix = ".".join(prefix.split(".")[1:])
    return dict(
        baseline_table=f"{prefix}_stable.baseline_v1",
        migration_table=f"{prefix}_stable.migration_v1",
        daily_table=f"{prefix}_derived.baseline_clients_daily_v1",
        last_seen_table=f"{prefix}_derived.baseline_clients_last_seen_v1",
        first_seen_table=f"{prefix}_derived.baseline_clients_first_seen_v1",
        daily_view=f"{prefix}.baseline_clients_daily",
        last_seen_view=f"{prefix}.baseline_clients_last_seen",
        first_seen_view=f"{prefix}.baseline_clients_first_seen",
        event_monitoring=f"{prefix}_derived.event_monitoring_live_v1",
        events_view=f"{prefix}.events",
        events_stream_table=f"{prefix}_derived.events_stream_v1",
        events_stream_view=f"{prefix}.events_stream",
    )


def referenced_table_exists(view_sql, id_token=None):
    """Dry run the given view SQL to see if its referent exists."""
    dryrun = DryRun("foo/bar/view.sql", content=view_sql, id_token=id_token)
    # 403 is returned if referenced dataset doesn't exist; we need to check that the 403 is due to dataset not existing
    # since dryruns on views will also return 403 due to the table CREATE
    # 404 is returned if referenced table or view doesn't exist
    return not any(
        [
            404 == e.get("code")
            or (
                403 == e.get("code")
                and "bigquery.tables.create denied" not in e.get("message")
            )
            for e in dryrun.errors()
        ]
    )


def _contains_glob(patterns):
    return any({"*", "?", "["}.intersection(pattern) for pattern in patterns)


def _extract_dataset_from_glob(pattern):
    # Assumes globs are in <dataset>.<table> form without a project specified.
    return pattern.split(".", 1)[0]


def get_app_info():
    """Return a list of applications from the probeinfo API."""
    resp = requests.get(APP_LISTINGS_URL)
    resp.raise_for_status()
    apps_json = resp.json()
    app_info = {}

    for app in apps_json:
        if app["app_id"].startswith("rally"):
            pass
        elif app["app_name"] not in app_info:
            app_info[app["app_name"]] = [app]
        else:
            app_info[app["app_name"]].append(app)

    return app_info


class GleanTable:
    """Represents a generated Glean table."""

    def __init__(self):
        """Init Glean table."""
        self.target_table_id = ""
        self.prefix = ""
        self.custom_render_kwargs = {}
        self.per_app_id_enabled = True
        self.per_app_enabled = True
        self.across_apps_enabled = True
        self.cross_channel_template = "cross_channel.view.sql"
        self.base_table_name = "baseline_v1"

    def skip_existing(self, output_dir="sql/", project_id="moz-fx-data-shared-prod"):
        """Existing files configured not to be overridden during generation."""
        return [
            file.replace(
                f'{ConfigLoader.get("default", "sql_dir", fallback="sql/")}{project_id}',
                str(output_dir),
            )
            for skip_existing in ConfigLoader.get(
                "generate", "glean_usage", "skip_existing", fallback=[]
            )
            for file in glob.glob(skip_existing, recursive=True)
        ]

    def generate_per_app_id(
        self,
        project_id,
        baseline_table,
        output_dir=None,
        use_cloud_function=True,
        app_info=[],
        parallelism=8,
        id_token=None,
    ):
        """Generate the baseline table query per app_id."""
        if not self.per_app_id_enabled:
            return

        tables = table_names_from_baseline(baseline_table, include_project_id=False)

        query_filename = f"{self.target_table_id}.query.sql"
        checks_filename = f"{self.target_table_id}.checks.sql"
        bigconfig_filename = f"{self.target_table_id}.bigconfig.yml"
        view_filename = f"{self.target_table_id[:-3]}.view.sql"
        view_metadata_filename = f"{self.target_table_id[:-3]}.metadata.yaml"
        table_metadata_filename = f"{self.target_table_id}.metadata.yaml"
        schema_filename = f"{self.target_table_id}.schema.yaml"

        table = tables[f"{self.prefix}_table"]
        view = tables[f"{self.prefix}_view"]
        derived_dataset = tables["daily_table"].split(".")[-2]
        dataset = derived_dataset.replace("_derived", "")

        app_name = dataset
        for app in app_info:
            for app_dataset in app:
                if app_dataset["bq_dataset_family"] == dataset:
                    app_name = app_dataset["app_name"]
                    break

        # Some apps did briefly send a baseline ping,
        # but do not do so actively anymore. This is why they get excluded.
        enable_monitoring = app_name not in list(set(BIGCONFIG_SKIP_APPS))

        print(f"App_name: {app_name} | monitoring: {enable_monitoring}")

        render_kwargs = dict(
            header="-- Generated via bigquery_etl.glean_usage\n",
            header_yaml="---\n# Generated via bigquery_etl.glean_usage\n",
            project_id=project_id,
            derived_dataset=derived_dataset,
            target_table=self.target_table_id,
            app_name=app_name,
            has_distribution_id=app_name in APPS_WITH_DISTRIBUTION_ID,
            has_profile_group_id=app_name in APPS_WITH_PROFILE_GROUP_ID,
            enable_monitoring=enable_monitoring,
        )

        render_kwargs.update(self.custom_render_kwargs)
        render_kwargs.update(tables)

        query_sql = render(
            query_filename, template_folder=PATH / "templates", **render_kwargs
        )
        view_sql = render(
            view_filename, template_folder=PATH / "templates", **render_kwargs
        )

        view_metadata = render(
            view_metadata_filename,
            template_folder=PATH / "templates",
            format=False,
            **render_kwargs,
        )
        table_metadata = render(
            table_metadata_filename,
            template_folder=PATH / "templates",
            format=False,
            **render_kwargs,
        )

        # Checks are optional, for now!
        try:
            checks_sql = render(
                checks_filename, template_folder=PATH / "templates", **render_kwargs
            )
        except TemplateNotFound:
            checks_sql = None

        # Schema files are optional
        try:
            schema = render(
                schema_filename,
                format=False,
                template_folder=PATH / "templates",
                **render_kwargs,
            )
        except TemplateNotFound:
            schema = None

        print(f"App_name: {app_name} | monitoring: {enable_monitoring}")

        if enable_monitoring:
            try:
                bigconfig_contents = render(
                    bigconfig_filename,
                    format=False,
                    template_folder=PATH / "templates",
                    **render_kwargs,
                )
            except TemplateNotFound:
                bigconfig_contents = None
        else:
            bigconfig_contents = None

        # generated files to update
        Artifact = namedtuple("Artifact", "table_id basename sql")
        artifacts = [
            Artifact(view, "metadata.yaml", view_metadata),
            Artifact(table, "metadata.yaml", table_metadata),
            Artifact(table, "query.sql", query_sql),
        ]

        if not (referenced_table_exists(view_sql, id_token)):
            logging.info("Skipping view for table which doesn't exist:" f" {table}")
        else:
            artifacts.append(Artifact(view, "view.sql", view_sql))

        skip_existing_artifact = self.skip_existing(output_dir, project_id)

        if output_dir:
            if checks_sql:
                if "baseline" in table and app_name in BQETL_CHECKS_SKIP_APPS:
                    logging.info(
                        "Skipped copying ETL check for %s as app: %s is marked as not having baseline ping"
                        % (table, app_name)
                    )
                else:
                    if checks_sql:
                        artifacts.append(Artifact(table, "checks.sql", checks_sql))

            if bigconfig_contents:
                artifacts.append(Artifact(table, "bigconfig.yml", bigconfig_contents))

            if schema:
                artifacts.append(Artifact(table, "schema.yaml", schema))

            for artifact in artifacts:
                destination = (
                    get_table_dir(output_dir, artifact.table_id) / artifact.basename
                )
                skip_existing = str(destination) in skip_existing_artifact

                write_sql(
                    output_dir,
                    artifact.table_id,
                    artifact.basename,
                    artifact.sql,
                    skip_existing=skip_existing,
                )

            write_dataset_metadata(output_dir, view)

    def generate_per_app(
        self,
        project_id,
        app_info,
        output_dir=None,
        use_cloud_function=True,
        parallelism=8,
        id_token=None,
    ):
        """Generate the baseline table query per app_name."""
        if not self.per_app_enabled:
            return

        app_name = app_info[0]["app_name"]

        target_view_name = "_".join(self.target_table_id.split("_")[:-1])
        target_dataset = app_name

        datasets = [
            (a["bq_dataset_family"], a.get("app_channel", "release")) for a in app_info
        ]

        if len(datasets) == 1 and target_dataset == datasets[0][0]:
            # This app only has a single channel, and the app_name
            # exactly matches the generated bq_dataset_family, so
            # the existing per-app_id dataset also serves as the
            # per-app dataset, thus we don't have to provision
            # union views.
            if self.per_app_id_enabled:
                return

        enable_monitoring = app_name not in list(
            set(BIGCONFIG_SKIP_APPS + BIGCONFIG_SKIP_APPS_METRICS)
        )

        render_kwargs = dict(
            header="-- Generated via bigquery_etl.glean_usage\n",
            header_yaml="---\n# Generated via bigquery_etl.glean_usage\n",
            project_id=project_id,
            target_view=f"{target_dataset}.{target_view_name}",
            datasets=datasets,
            table=target_view_name,
            target_table=f"{target_dataset}_derived.{self.target_table_id}",
            app_name=app_name,
            enable_monitoring=enable_monitoring,
        )
        render_kwargs.update(self.custom_render_kwargs)

        skip_existing_artifacts = self.skip_existing(output_dir, project_id)

        Artifact = namedtuple("Artifact", "table_id basename sql")

        if self.cross_channel_template:
            sql = render(
                self.cross_channel_template,
                template_folder=PATH / "templates",
                **render_kwargs,
            )
            view = f"{project_id}.{target_dataset}.{target_view_name}"

            if not (referenced_table_exists(sql, id_token=id_token)):
                logging.info("Skipping view for table which doesn't exist:" f" {view}")
                return

            if output_dir:
                write_dataset_metadata(output_dir, view)

            if output_dir:
                skip_existing = (
                    str(get_table_dir(output_dir, view) / "view.sql")
                    in skip_existing_artifacts
                )
                write_sql(
                    output_dir, view, "view.sql", sql, skip_existing=skip_existing
                )
        else:
            query_filename = f"{target_view_name}.query.sql"
            query_sql = render(
                query_filename, template_folder=PATH / "templates", **render_kwargs
            )
            view_sql = render(
                f"{target_view_name}.view.sql",
                template_folder=PATH / "templates",
                **render_kwargs,
            )
            metadata = render(
                f"{self.target_table_id[:-3]}.metadata.yaml",
                template_folder=PATH / "templates",
                format=False,
                **render_kwargs,
            )

            table = f"{project_id}.{target_dataset}_derived.{self.target_table_id}"
            view = f"{project_id}.{target_dataset}.{target_view_name}"

            if not (referenced_table_exists(query_sql, id_token=id_token)):
                logging.info(
                    "Skipping query for table which doesn't exist:"
                    f" {self.target_table_id}"
                )
                return

            if output_dir:
                artifacts = [
                    Artifact(table, "query.sql", query_sql),
                    Artifact(table, "metadata.yaml", metadata),
                    Artifact(view, "view.sql", view_sql),
                ]

                if (
                    self.target_table_id.startswith("metrics_clients_")
                    and enable_monitoring
                ):
                    bigconfig_contents = render(
                        f"{self.target_table_id[:-3]}.bigconfig.yml",
                        format=False,
                        template_folder=PATH / "templates",
                        **render_kwargs,
                    )

                    artifacts.append(
                        Artifact(table, "bigconfig.yml", bigconfig_contents)
                    )

                for artifact in artifacts:
                    destination = (
                        get_table_dir(output_dir, artifact.table_id) / artifact.basename
                    )
                    skip_existing = destination in skip_existing_artifacts

                    write_sql(
                        output_dir,
                        artifact.table_id,
                        artifact.basename,
                        artifact.sql,
                        skip_existing=skip_existing,
                    )

                write_dataset_metadata(output_dir, view)
                write_dataset_metadata(output_dir, table, derived_dataset_metadata=True)

    def generate_across_apps(
        self, project_id, apps, output_dir=None, use_cloud_function=True, parallelism=8
    ):
        """Generate a query across all apps."""
        # logic for implementing cross-app queries needs to be implemented in the
        # individual classes
        return
